from typing import TypeVar,List
A = TypeVar('A')

def flatten(lol : List[List[A]]) -> List[A]:
    """
    Convert list of lists to a single list via concatenation
    """
    return [item for sublist in lol for item in sublist]



from typing import Tuple,List
from json import loads
from dbgen.core.lists import flatten

def get_catalog_ads(jsonads:str)->Tuple[List[int],List[int]]:
    """
    Analyzes params.json's "adsorbates" field
    """
    if not jsonads:
        return ([],[])
    else:
        ads = loads(jsonads)
        if not ads or ads==[[]]:
            return ([],[])
        output = [[(i,a) for a in ad] for i,ad in enumerate(ads)]
        try:
            ads,inds = zip(*flatten(output))
        except:
            import pdb;pdb.set_trace()
        return list(ads),list(inds)



from networkx import node_link_graph,MultiGraph # type: ignore
from json import loads

def json_to_graph(raw : str) -> MultiGraph:
    """
    Deserialize a MultiGraph
    """
    return node_link_graph(loads(raw))

if __name__=='__main__':
    from dbgen.main import realDB
    from dbgen.support.utils import sqlselect
    import sys
    q = """SELECT geo_graph
            FROM alljob JOIN finaltraj USING (job_id)
                        JOIN struct USING (struct_id)
            WHERE  stordir = %s"""
    gstr = sqlselect(realDB,q,[sys.argv[1]])[0][0]
    G = json_to_graph(gstr)



# External Modules
from typing import Optional,Any,List,Callable,TYPE_CHECKING
if TYPE_CHECKING:
    from dbgen.support.datatypes.rule import Rule

from MySQLdb import connect,Connection,OperationalError             # type: ignore
from MySQLdb.cursors import Cursor,DictCursor                       # type: ignore
from time import sleep
from os import environ
from random import random
from pprint import pformat
################################################################################

class ExternalError(Exception):
    """
    Custom class for catching errors that occur in code external to dbgen
    """
    def __init__(self, message : str) -> None:
        super().__init__(message)

################################################################################
localuser = environ["USER"]

class ConnectInfo(object):
    """MySQL connection info """
    def __init__(self
                ,host   : str = '127.0.0.1'
                ,port   : int = 3306
                ,user   : str = localuser
                ,passwd : str = localuser
                ,db     : str = 'suncat'
                ) -> None:
        self.host   = host
        self.port   = port
        self.user   = user
        self.passwd = passwd
        self.db     = db

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def mk_conn(self
               ,mk_dict : bool = False
               ,attempt : int  = 10
               ) -> Connection:
        try:
            con = connect(host   = self.host
                          ,port   = self.port
                          ,user   = self.user
                          ,passwd = self.passwd
                          ,db     = self.db
                          ,cursorclass = DictCursor if mk_dict else Cursor
                          ,connect_timeout=28800)
                          #,wait_timeout = 28800
                          #,interactive_timeout = 28800)

        except OperationalError as e:
            if attempt > 0:
                sleep(5)
                con = self.mk_conn(mk_dict,attempt-1)
            else:
                raise OperationalError(e)
        con.query('SET SESSION wait_timeout=28800')
        return con
################################################################################
class Arg(object):
    """
    How a function refers to a namespace
    """

    def __init__(self
                ,name : str
                ,ind  : Optional[int] = None
                ) -> None:
        self.name = name
        self.ind = ind

    def src(self)->str:
        ind = '' if self.ind is None else ',%s'%self.ind
        return "Arg('%s'%s)"%(self.name,ind)
    def arg_get(self,dic:dict)->Any:
        try:
            val = dic[self.name]
            if self.ind is None:
                return val
            else:
                try:
                    return val[self.ind]
                except IndexError:
                    args= [self.name,self.ind,val,pformat(dic)]
                    msg = 'cannot find {0}''s index {1} in {2} \ndic = {3}'
                    raise IndexError(msg.format(*args))
                except TypeError:
                    raise TypeError('cannot index into %s: '%val.__class__,val)
        except KeyError:
            raise KeyError('could not find %s in '%self.name,list(dic.keys()))

    def __str__(self)->str:
        suffix = '' if self.ind is None else '[%d]'%self.ind
        return 'Arg(%s%s)'%(self.name,suffix)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def add_arg(self
               ,sqlite_pth : str
               ,arg_kind   : str
               ,trans_id   : int
               ,next_id    : int
               ,arg_id     : int
               ) -> None:
        from dbgen.support.utils import mkInsCmd,sqlite_execute
        if arg_kind == 'meta':
            tab = 'blockarg'
        elif arg_kind == 'temp':
            tab = 'tempfuncarg'
        else:
            raise NotImplementedError

        id_col = '%s_id'%(tab[:-3]) # blockarg -> block_id
        q = mkInsCmd(tab,['rule_id',id_col,tab+'_id','name','ind'],sqlite=True)
        sqlite_execute(sqlite_pth,q,[trans_id,next_id,arg_id,self.name,self.ind])

def noIndex(xs:List[str])->List[Arg]:
    """
    At some point we represented arguments as a dictionary with namespace
    elements as keys and an optional index as values. Frequently we don't want
    to index, so this function simplifies writing the values.
    """
    return [Arg(x,None) for x in xs]

###############

class Test(object):
    """
    Execute a test before running rule. If it returns True, the test is
    passed, otherwise it returns an object which is fed into the "message"
    function. This prints a message: "Not Executed (<string of object>)"
    """
    def __init__(self
                ,test    : Callable[['Rule',Any],Any]
                ,message : Callable[[Any],str]
                ) -> None:
        self.test = test
        self.message = message

    def run_test(self,t:'Rule',*args:Any)->Any:
        output = self.test(t,*args)
        if output is True:
            return True
        else:
            return self.message(output)



onlyTest = Test(lambda t,o: (len(o) == 0) or (t.name in o) # type: ignore
               ,lambda x: "Rule not in 'Only' input specification")
xTest = Test(lambda t,x:  t.name not in x # type: ignore
            ,lambda x: "Rule in 'Exclude' input specification")
fTest = Test(lambda t,conn: t.dependency_failed(conn)
            ,lambda x: "Rule depends on failed jobs %s"%pformat(x))



from typing import Tuple

import json

def get_cell(json_atoms : str)->Tuple[float,float,float,float,float,float,float,float,float]:
    """
    Extracts information about a list of atoms (in json form) and prepares cell
    info for insertion into cell table
    """
    atoms = json.loads(json_atoms)
    [a,b,c] = atoms['cell']
    return  tuple(a+b+c) #type: ignore



from typing import Dict
from dbgen.core.parsing import parse_line

def get_dipole_vasp(outcar:str) -> int:
    """
    Parse INCAR to detect if dipole correction was applied
    """
    parsed = parse_line(outcar,'LDIPOL',0)
    if parsed is None:
        raise ValueError('malformed OUTCAR? no "LDIPOL" found')
    else:
        return int(' T ' in parsed)



from typing import Tuple,Dict,Optional
import json
from dbgen.core.parsing import parse_line

def get_nbands_qe(files_params  : Tuple[str,str]) -> Optional[float]:
    """
    If we do have a params json, try to get nbands from it.
    """
    params_raw = files_params[1]
    if not params_raw:
        return None
    else:
        params = json.loads(params_raw)
        return params.get('nbands')     # no conceivable way of getting this from log



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_diag_qe(log : str) -> Optional[str]:
    """
    Diagonalization algorithm
    """
    if 'Davidson diagonalization' in log:
        return 'david'
    else:
        parsed = parse_line(log,'diagonalization',0)
        if parsed is None:
            return None
        else:
            if 'Davidson' in parsed:
                return 'david'
            else:
                raise NotImplementedError('new diag routine in QE? '+parsed)



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_sigma_gpaw(log:str) -> Optional[float]:
    """
    Helpful docstring
    """
    parsed = parse_line(log,'Fermi-Dirac',0)
    if parsed is None:
        raise ValueError
    else:
        raw = parsed.split('=')[1].split()[0]
        return float(raw)



from typing import Dict
from dbgen.core.parsing import parse_line

def get_diag_gpaw(log : str ) ->str:
    """
    Diagonalization algorithm.
    """
    parsed    = parse_line(log,'eigensolver',0)
    if parsed is None:
        raise ValueError('Malformed GPAW log file? can''t find "eigensolver"')
    else:
        eigenline = parsed.split(': ')[-1]
        if 'dav' in eigenline:
            return 'david'
        else:
            return eigenline



#External Modules
from typing import Callable,List,Optional,Union,Dict
#Internal Modules
from dbgen.support.datatypes.select         import Select
from dbgen.support.datatypes.temp           import Temp,TempFunc,Const
from dbgen.support.datatypes.insertupdate   import InsertUpdate

from dbgen.support.datatypes.rule       import Rule
from dbgen.support.datatypes.metatemp   import MetaTemp
from dbgen.support.datatypes.misc       import Arg

################################################################################

class Relation(object):
    """
    Relate attributes of objects via triples of:
    -select:  Extracting info from the current state of the world
    -func:    User-defined data-processing pipeline
    -inserts: putting information back into the DB
    """
    def __init__(self
                ,name      : str
                ,select    : Optional[Select]   = None
                ,func      : Temp               = Temp()
                ,inserts   : List[InsertUpdate] = []
                ) -> None:
        self.name    = name
        self.select  = select
        self.func    = func
        self.inserts = inserts
    def mk_rule(self)->Rule:
        return Rule(name     = self.name
                   ,query    = str(self.select) if self.select else None
                   ,template = self.func
                   ,metatemp = MetaTemp([iu.mk_block() for iu in self.inserts]))



from typing import TypeVar,Callable,List
A = TypeVar('A')
B = TypeVar('B')

from dbgen.core.lists import flatten


def concat_map(f : Callable[[A],List[B]]
              ,args:List[A]
              ) -> List[B]:
    """
    Maps a function over an input. Function is expected to return ([a],[b],...)
    We apply the function to every element in the list and concatenate each tuple
    element.

    """
    mapped = [f(arg) for arg in args]

    if mapped == []:
        return []
    elif isinstance(mapped[0],list):
        return flatten(mapped)
    elif isinstance(mapped[0],tuple):
        example = list(zip(*mapped))
        return tuple(map(flatten,zip(*mapped)))
    else:
        raise ValueError('did you mean to use concat_map?',mapped)



from typing import Any,TYPE_CHECKING,Set
from abc    import abstractmethod
if TYPE_CHECKING:
    from dbgen.support.datatypes.attr import Attr
    # from dbgen.support.datatypes.object import _Object

################################################################################
class Constraint(object):
    def __init__(self)->None: pass

    def __repr__(self)->str:
        return 'Constraint(%s)'%(str(self))

    @abstractmethod
    def attrs(self)->Set['Attr']:
        raise NotImplementedError
################################################################################
class UnaryConstraint(Constraint):
    def __init__(self,obj:Any)->None:
        self.obj = obj
    def attrs(self)->Set['Attr']:
        """
        Avoid Cyclic
        """
        from dbgen.support.datatypes.object import Attr
        return set([self.obj]) if isinstance(self.obj,Attr) else set()
class NOT(UnaryConstraint):
    def __init__(self,obj:Any)->None:
        super().__init__(obj)
    def __str__(self)->str:
        return "NOT %s"%self.obj

class NULL(UnaryConstraint):
    def __init__(self,obj:Any)->None:
        super().__init__(obj)
    def __str__(self)->str:
        return "%s IS NULL"%self.obj

class IN(UnaryConstraint):
    def __init__(self,obj:Any,lst:list)->None:
        super().__init__(obj)
        self.lst = lst
    def __str__(self)->str:
        return "%s IN (%s)"%(self.obj,','.join(map(str,self.lst)))

################################################################################
class BinaryConstraint(Constraint):
    def __init__(self,o1:Any,o2:Any)->None:
        self.o1 = o1
        self.o2 = o2

    def attrs(self)->Set['Attr']:
        from dbgen.support.datatypes.object import Attr
        s1 = set([self.o1]) if isinstance(self.o1,Attr) else set()
        s2 = set([self.o2]) if isinstance(self.o2,Attr) else set()
        return s1 | s2

class EQ(BinaryConstraint):
    def __init__(self,o1:Any,o2:Any)->None:
        super().__init__(o1,o2)

    def __str__(self)->str:
        return '%s = %s'%(self.o1,self.o2)

class NE(BinaryConstraint):
    def __init__(self,o1:Any,o2:Any)->None:
        super().__init__(o1,o2)

    def __str__(self)->str:
        return '%s != %s'%(self.o1,self.o2)

class LT(BinaryConstraint):
    def __init__(self,o1:Any,o2:Any)->None:
        super().__init__(o1,o2)

    def __str__(self)->str:
        return '%s < %s'%(self.o1,self.o2)

class GT(BinaryConstraint):
    def __init__(self,o1:Any,o2:Any)->None:
        super().__init__(o1,o2)

    def __str__(self)->str:
        return '%s > %s'%(self.o1,self.o2)
# #####################
# def HAS(o:'_Object',comp:'_Object')->UnaryConstraint:
#     return NOT(HASNT(o,comp))
#
# def HASNT(o:'_Object',comp:'_Object')->UnaryConstraint:
#     for fk in o.fks:
#         if fk._to == comp:
#             return NULL(fk.from_attr[0])
#     raise ValueError



from typing import Dict
from dbgen.core.parsing import parse_line

def get_dipole_gpaw(log:str) -> int:
    """
    Until we figure out how to actually get dipole correction to work,
    assume that it's always turned off
    """

    return 0



from typing import Dict
from dbgen.core.parsing import parse_line

def parse_pw_vasp(outcar:str) -> int:
    """
    Docstring
    """
    parsed  = parse_line(outcar,'ENCUT',0)
    if parsed is None:
        raise ValueError
    else:
        raw_num = parsed.split('=')[-1].split()[0]

    return int(round(float(raw_num)))



# External Modules
from typing     import List,Tuple,Optional
from ase        import Atoms # type: ignore
from string     import digits
from os.path    import join,exists
from os         import listdir
# Internal Modules
from dbgen.core.numeric                     import roundfloat
from dbgen.scripts.Pure.Atoms.json_to_traj  import json_to_traj

ForceTuple = Tuple[int,int,Optional[float],Optional[float],Optional[float]]
################################################################################
def get_traj_qe_lazy(a_l_p : Tuple[str,str,str])->List[Tuple[Atoms,int,float,List[ForceTuple],int]]:
    """
    Returns list of triples to represent the trajectory of a structure during
    an optimization. The ASE structure, the step number along the trajectory,
    and the energy. These are parsed from a Quantum Espresso log file.
    """
    anytraj,log,_ =a_l_p
    # Initialize
    #-----------
    output   = []
    CELL     = False
    POS      = False
    FORCE    = False
    cellvecs = [] # contains up to three entries like [0.000,1.15,16.324]
    atomdata = [] # contains many (SYMB,X,Y,Z) tuples
    forces   = []
    stepnum  = 0  # step in the trajectory
    traj  = json_to_traj(anytraj)

    # Constants
    #----------
    ryd_to_ev = 13.60569

    # Get log file
    #-------------
    lines = log.split('\n')
    for line in lines: # read lazily

        # MAIN LOOP
        #----------
        if 'crystal axes' in line:
            CELL = True
            cellvecs = []
        elif CELL and len(cellvecs)<3:
            try:
                cellvecs.append([roundfloat(x) for x in line.split()[3:6]])
            except ValueError:
                CELL = False
                cellvecs = []
            if 'a(3)' in line:
                CELL = False
        elif 'site n.' in line:
            POS = True
            atomdata = []
        elif 'number of atoms/cell' in line:
            num_atoms = int(line.split()[-1])
        elif POS:
            try:
                ind,sym,_,_,_,_,x,y,z,_ = line.split()
                remove_digits = str.maketrans('', '', digits)
                clean_symb = sym.translate(remove_digits)
                atomdata.append((clean_symb,(roundfloat(x),roundfloat(y),roundfloat(z))))
                if int(ind)== num_atoms:
                    POS = False
                    if cellvecs: # construct atoms if cell isn't malformed
                        atoms = Atoms(symbols = [a[0] for a in atomdata]
                                     ,positions = [a[1] for a in atomdata]
                                     ,cell = cellvecs
                                     ,constraint=traj.constraints
                                     ,tags=traj.get_tags())
                        atoms.wrap()
            except ValueError:
                POS = False
                atomdata = []
        elif 'total energy      ' in line:
            eng = float(line.split()[-2])
        elif  'smearing contrib' in line:
            ts = float(line.split()[-2])

        elif 'Forces acting' in line:
            FORCE = True
            forces = []

        elif FORCE and 'Total force' in line:
            FORCE = False
        elif FORCE and line.strip()!='' and 'rho' not in line:
            try:
                _,ind1,_,_,_,_,fx,fy,fz = line.split()
                forces.append((stepnum,int(ind1)-1,roundfloat(float(fx)* ryd_to_ev)
                                                  ,roundfloat(float(fy)* ryd_to_ev)
                                                  ,roundfloat(float(fz)* ryd_to_ev)))
            except ValueError:
                FORCE=False
                forces = []
        elif 'JOB DONE' in line:
            e = (eng-ts/2) * ryd_to_ev

            if all([cellvecs, atomdata , forces
                   , len(traj)==num_atoms
                   , len(traj)==len(forces)]):
                output.append((traj,stepnum,e,forces,0))
                stepnum += 1
                e = 0

    if forces: # premature end of file before job done
        e = (eng-ts/2) * ryd_to_ev
        output.append((atoms,stepnum,e,forces,1))


    output[-1] = tuple([*output[-1][:-1],1]) # set final for finaltraj
    return output



import json
import ase                              # type: ignore
import  pymatgen as pmg                 # type: ignore
import  pymatgen.symmetry.analyzer as pysym  # type: ignore

def get_pointgroup(atoms : ase.Atoms) -> str:
	"""
    uses pymatgen.symmetry.analyzer
	"""
	atoms.center()
	symbs = atoms.get_chemical_symbols()
	pos   = atoms.get_positions()
	mol   = pmg.Molecule(symbs,pos)
	return pysym.PointGroupAnalyzer(mol).sch_symbol



from typing import TypeVar, Tuple

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')

def triple(a : A
          ,b : B
          ,c : C
          ) -> Tuple[A,B,C] :
    """
    Three items into a tuple
    """
    return (a,b,c)



from pymatgen import Structure  # type: ignore
from ase import Atoms  # type: ignore
from pymatgen.io.ase            import AseAtomsAdaptor      # type: ignore

def pmg_to_ase(pmg:Structure)->Atoms:
    """
    Convert a Pymatgen structure into an ASE atoms object
    """
    return AseAtomsAdaptor().get_atoms(pmg)



# External modules
from ase        import Atoms    # type: ignore
from ase.io     import read     # type: ignore
from os.path    import join,exists,getsize
from os         import environ,mkdir,listdir
from subprocess import call
from shutil     import rmtree
from string     import ascii_uppercase,digits
from random     import choices

# Internal modules
from dbgen.scripts.Pure.Atoms.traj_to_json import traj_to_json
################################################################################
def anytraj(root : str) -> str:
    """
    ASE IO read function - takes any traj it can find
    """
    user = environ['SHERLOCK_USERNAME']
    rand = ''.join(choices(ascii_uppercase+digits, k=16))
    mkdir(rand)

    call(['scp','-C','-r','-q','%s@login.sherlock.stanford.edu:%s/*.traj'%(user,root)
         ,rand])
    trajs = [join(rand,t) for t in listdir(rand) if getsize(join(rand,t)) > 100]

    if len(trajs) == 0:
        raise ValueError('Get Traj could not find any traj in '+root)
    else:
        atoms = read(trajs[0])
        rmtree(rand)
        return traj_to_json(atoms)

if __name__=='__main__':
    import sys
    print(anytraj(sys.argv[1]))



import ase # type: ignore
import json
#from numpy.linalg import norm  # type: ignore

from dbgen.core.numeric import roundfloat
def traj_to_json(atoms : ase.Atoms) -> str:
    """
    Serialize an Atoms object in a human readable way
    """

    atomdata = []

    fixed_inds = []
    if atoms.constraints:
        for constraint in atoms.constraints:
            if isinstance(constraint,ase.constraints.FixAtoms):
                fixed_inds.extend(list(constraint.get_indices()))

    atoms.wrap()
    for a in atoms: atomdata.append({'number'       : int(a.number)
                                    ,'x'            : roundfloat(a.x)
                                    ,'y'            : roundfloat(a.y)
                                    ,'z'            : roundfloat(a.z)
                                    ,'magmom'       : roundfloat(a.magmom)
                                    ,'tag'          : int(a.tag)
                                    ,'constrained'  : int(a.index in fixed_inds)
                                    ,'index'        : int(a.index)})

    out = {'cell': [[roundfloat(x) for x in xx] for xx in atoms.get_cell().tolist()]
          ,'atomdata':atomdata}

    return json.dumps(out)



from typing import  Tuple,List,Dict

# External Modules
import numpy as np  # type: ignore
from os.path     import join
from collections import defaultdict
from ase         import Atoms,Atom
# Internal Modules
from dbgen.core.lists           import flatten
from dbgen.scripts.IO.get_file  import get_file

################################################################################

def parse_chargemol(stordir : str
                   ,atoms   : Atoms
                   ) -> Tuple[List[int]
                                 ,List[int]
                                 ,List[int]
                                 ,List[int]
                                 ,List[int]
                                 ,List[float]
                                 ,List[int]
                                 ,List[float]]:
    """
    Analyzes the contents of a chargemol calculation directory.

    Returns an unzipped list of <ind,ind,bondorder> triples
    followed by an unzipped list of <ind,charge> pairs

    """

    # Auxillary Functions
    #--------------------
    def parse_line(line : str) -> Tuple[int,float,List[int]]:
        """
        Get bonded atom, bond order, and offset
        """
        assert line[:16]==' Bonded to the (', "No parse line ->"+line+'<- %d'%len(line)
        offStr,indStr,boStr = line[16:].split(')')
        offset = [int(x) for x in offStr.split(',')] # Chunk containing offset info
        ind    = int(indStr.split()[-3]) - 1        # Chunk containing index info (chargemol starts at 1, not 0)
        bo     = float(boStr.split()[4])            # Chunk containing B.O. info
        return (ind,bo,offset)

    def mk_pbc_dict(atoms : Atoms
                   ) -> Dict[int,Tuple[int,int,int]]:
        """
        Helpful docstring
        """
        def g(tup : tuple)->Tuple[int,int,int]:
            """
            Helper function to yield tuples for pbc_dict
            """
            def f(x : float)->int:
                """
                Helper function for g
                """
                if   x < 0: return -1
                elif x < 1: return 0
                else:       return 1

            x,y,z = tup
            return (f(x),f(y),f(z))

        scaled_pos  = atoms.get_scaled_positions(wrap=False).tolist()
        scaled_pos_ = zip(range(len(atoms)),scaled_pos)
        pbc_dict    = {i : g(p) for i,p in scaled_pos_}
        return pbc_dict

    def dict_diff(d1 : Dict[int,np.array]
                 ,d2 : Dict[int,np.array]
                 ) -> Dict[int,np.array]:
        """
        Helpful docstring
        """
        return {i: np.array(d2[i]) - np.array(d1[i]) for i in d1.keys()}


    # Handle sortdict
    #----------------
    try:
        ase_sort = get_file(join(stordir,'ase-sort.dat'))
        loglines = map(lambda x: map(int,x.split()),ase_sort.split('\n'))
        sortdict = {x:y for x,y in loglines} # type: Dict[int,int]
    except IOError:
        sortdict = {}

    # Classes
    #--------

    class PotentialEdge(object):
        """
        Container for information we need to decide later if it's a
        graph-worthy bond. This class is nothing more than a dictionary.
        """
        def __init__(self
                    ,fromNode   : int
                    ,toNode     : int
                    ,bond_id    : int
                    ,distance   : float
                    ,offset     : List[int]
                    ,bond_order : float
                    ) -> None:

            self.fromNode = fromNode
            self.toNode   = toNode
            self.bond_id  = bond_id
            self.pbc_x    = offset[0]
            self.pbc_y    = offset[1]
            self.pbc_z    = offset[2]
            self.distance = distance
            self.bondorder= bond_order


    class BondOrderSection(object):
        """
        Process one section of the Bond Order output of Chargemol
        """
        def __init__(self
                    ,ind       : int
                    ,raw_lines : List[str]
                    ,pbcdict   : dict
                    ) -> None:

            self.ind      = ind
            self.bonds    = [parse_line(x) for x in raw_lines]
            self.pbcdict  = pbcdict

            self.bondcounter = defaultdict(int) # type: Dict[Tuple[int,int],int]

        def _relative_shift(self
                           ,i : int
                           ,j : int
                           ) -> np.array:
            """
            Given a pbc_dict and two indices, return the original pbc shift
            for a bond from i to j
            """
            pi,pj = [np.array(self.pbcdict[x]) for x in [i,j]]
            return pj - pi

        def makeEdge(self
                    ,tup :  Tuple[int,float,List[int]]
                    ) -> PotentialEdge:
            """
            Creates an Edge instance from the result of a parsed Bond Order log line
            """
            (toInd,bo,offset) = tup
            fromInd = self.ind

            if sortdict: # undo VASP reordering if necessary
                fromInd,toInd = [sortdict[x] for x in [fromInd,toInd]] # type: ignore

            # correct for WRAPPING atoms
            offset = (np.array(offset) +  self._relative_shift(fromInd,toInd)).tolist()

            shift = np.dot(offset, atoms.get_cell()) # PBC shift
            p1    = atoms[fromInd].position
            p2    = atoms[toInd].position + shift
            d     = np.linalg.norm(p2-p1)

            #

            bondid = self.bondcounter[(fromInd,toInd)]
            self.bondcounter[(fromInd,toInd)] += 1
            return PotentialEdge(fromInd,toInd,bondid,d,offset,bo)

        def make_edges(self)->List[PotentialEdge]:
            """Apply edgemaker to result of parsing logfile lines"""
            return  [self.makeEdge(b) for b in self.bonds]

    def parse_chargemol_pbc(header_lines : List[str]
                           ,cell         : List[float]
                           ) -> Dict[int,Tuple[int,int,int]]:
        """
        Helpful docstring
        """
        atoms = Atoms(cell=cell)

        for i,l in enumerate(header_lines[2:]):
            try:
                if '*' in l:
                    l = l.replace('*','') + ' ???' #Asterisks are touching the z value
                s,x,y,z,_ = l.split()
                p = [float(q) for q in [x,y,z]]
                atoms.append(Atom(s,position=p))
            except Exception as e:
                pass #print('exception???',e)

        return mk_pbc_dict(atoms)

    # Main Program
    #-------------
    header,content    = [], [] # type: Tuple[List[str],List[str]]
    sections          = []     # type: List[BondOrderSection]
    filtedges         = []
    head_flag,counter = True,-1       # Initialize
    filepath  = join(stordir,'DDEC6_even_tempered_bond_orders.xyz')# File to parse
    bondlines = get_file(filepath).split('\n')
    pbcdict   = mk_pbc_dict(atoms) # the 'true' PBC coordinate of each atom

    for line in bondlines:
        if not (line.isspace() or line[:4] == ' 201'): # remove blank lines and calendar date
            if line[1]=='=':
                head_flag = False                                            # no longer in header  section
                chargemol_pbc = parse_chargemol_pbc(header,atoms.get_cell()) # PBC of chargemol atoms

            elif head_flag: header.append(line)     # we're still in the header section
            elif 'Printing' in line:                # we're describing a new atom
                content = []                        # reset buffer content
                counter += 1                        # update index of our from_index atom
            elif 'sum' in line:                     # summary line at end of a section
                dic = dict_diff(pbcdict,chargemol_pbc)
                edgs = BondOrderSection(counter,content,dic).make_edges()
                filtedges.extend([e for e in edgs
                    if  (e.bondorder > 0.04) and (e.fromNode <= e.toNode)])
            else:
                content.append(line)              # business-as-usual, add to buffer

    i1s,i2s,bondids,xs,ys,zs,dists,bos = map(list,zip(*[e.__dict__.values() for e in filtedges]))


    # Get charges
    #------------

    fpath     = join(stordir,'DDEC6_even_tempered_net_atomic_charges.xyz')
    chg_lines = get_file(fpath).split('\n')

    head_flag = True
    tail_flag = False
    chgs      = [] # List[float]

    for line in chg_lines[:-3]:
        if 'eigenvalues' in line:
            head_flag = False # we are no longer in the header region
        if not head_flag and not tail_flag:
            if line.isspace():
                tail_flag = True # we are now in the tail region
            else:
                try: # we are in the region with charges, for the most part
                    chgs.append(float(line.split()[5])) # first line bogus
                except ValueError:
                    pass
    inds = list(range(len(atoms)))

    if sortdict: # undo VASP reordering if necessary
        charges = [chgs[sortdict[i]] for i in inds] # type: ignore
    else:
        charges = chgs
    return (i1s,i2s,bondids,xs,ys,zs,dists,bos,inds,charges) # type: ignore


if __name__=='__main__':
    from dbgen.core.misc import anytraj
    import sys
    pth = sys.argv[1]
    print(parse_chargemol(pth,anytraj(pth)))



# External Modules
from typing  import Dict,List,Tuple,Optional
from ase     import Atoms # type: ignore
from os.path import join,exists
# Internal Modules
from dbgen.core.numeric import roundfloat
from dbgen.core.lists   import flatten
from dbgen.core.misc    import wrap
from dbgen.scripts.Pure.Atoms.json_to_traj   import json_to_traj

ForceTuple = Tuple[int,int,Optional[float],Optional[float],Optional[float]]
################################################################################
def get_traj_vasp_lazy(a_l_p : Tuple[str,str,str])->List[Tuple[Atoms,int,float,List[ForceTuple],int]]:
    """
    Returns list of quadruples to represent the trajectory of a structure during
    an optimization.
        - ASE structure,
        - the step number along the trajectory,
        - the energy
        - the forces.

    These are parsed from VASP's OUTCAR and POSCAR.
    """
    # Initialize
    #-----------
    anytraj,outcar,poscar = a_l_p
    output   = []
    CELL     = False
    POSFORCE = False
    cellvecs = [] # contains up to three entries like [0.000,1.15,16.324]
    pos = [] # contains many (X,Y,Z) tuples
    forces   = []
    stepnum  = 0  # step in the trajectory
    atomcounter = 0
    traj  = json_to_traj(anytraj)

    # Prequel: extract the list of chemical symbols from POSCAR in a convoluted way
    #-----------------------------------------------------------------------------
    plines = poscar.split('\n')
    uniqsymbs = plines[0].split() # ['H','O','Na']
    quantities = plines[5].split() # ['2','1','5']
    symbs = flatten([[a]*int(b) for a,b in zip(uniqsymbs,quantities)]) # ['H','H','O','Na','Na','Na','Na','Na']

    # Get log file
    #-------------
    with open(outcar,'r') as f:
        for line in f: # read lazily

            # MAIN LOOP
            #----------
            if 'reciprocal lattice vectors' in line:
                CELL = True
                cellvecs = []
            elif CELL and len(cellvecs)<3:
                cellvecs.append([roundfloat(x) for x in line.split()[:3]])
            elif CELL and 'length of vectors' in line:
                CELL = False

            elif 'POSITION' in line:
                POSFORCE = True
            elif POSFORCE and '---' not in line:
                if 'total drift' in line:
                    POSFORCE = False
                else:
                    x,y,z,fx,fy,fz = line.split()
                    pos.append([roundfloat(x),roundfloat(y),roundfloat(z)])
                    forces.append((stepnum,atomcounter,fx,fy,fz))
                    atomcounter +=1
            elif "free  energy   TOTEN" in line:
                e = float(line.split()[-2])
                atoms = Atoms(symbols=symbs,positions=pos,cell=cellvecs
                             ,constraint=traj.constraints,tags=traj.get_tags())
                wrap(atoms)
                for a in atoms:
                    assert a.z >= 0
                output.append((atoms,stepnum,e,forces,0))
                atomcounter = 0
                cellvecs,pos,forces = [],[],[]
                stepnum  += 1   # step in the trajectory


    output[-1] = tuple([*output[-1][:-1],1]) # set final for finaltraj
    return output



from typing import Dict,Optional,Tuple
import json
from dbgen.core.parsing import parse_line

def get_nbands_vasp(files_params : Tuple[str,str]) -> Optional[float]:
    """
    docstring
    """
    outcar = files_params[0]
    parsed = parse_line(outcar,'NBANDS',0)
    if parsed is None:
        raise ValueError('Malformed OUTCAR? No "NBANDS" found')
    else:
        nbands = parsed.split('=')[-1]
        return int(nbands)



from typing import Tuple
import json
import ase # type: ignore


def get_kptden(kx : int
              ,ky : int
              ,kz : int
              ,ax : float
              ,ay : float
              ,az : float
              ,bx : float
              ,by : float
              ,bz : float
              ,cx : float
              ,cy : float
              ,cz : float
              ) -> Tuple[float,float,float]:
    """
    Returns kpt density:
        https://wiki.physics.udel.edu/phys824/About_k-point_sampling
    """

    atoms = ase.Atoms(numbers   = [1] # fake Atoms object with correct cell
                     ,positions = [[0,0,0]]
                     ,cell      = [[ax,ay,az],[bx,by,bz],[cx,cy,cz]])

    recipcell = atoms.get_reciprocal_cell()

    denoms = [2. * 3.1415 * (recipcell[i]**2).sum()**(0.5) for i in range(3)]

    kptdenx = round(kx/denoms[0],2)
    kptdeny = round(ky/denoms[1],2)
    kptdenz = round(kz/denoms[2],2)

    return kptdenx,kptdeny,kptdenz



from typing import Dict,List,Tuple,Optional
from re import compile,VERBOSE
from ase import Atoms # type: ignore
from os.path import join,exists

from dbgen.core.parsing import parse_line,btw
from dbgen.core.numeric import roundfloat
from dbgen.core.lists   import flatten


ForceTuple = Tuple[int,int,Optional[float],Optional[float],Optional[float]]

def get_traj_vasp(stordir : str)->List[Tuple[Atoms,int,float,List[ForceTuple],int]]:
    """
    Returns list of quadruples to represent the trajectory of a structure during
    an optimization.
        - ASE structure,
        - the step number along the trajectory,
        - the energy
        - the forces.

    These are parsed from VASP's OUTCAR and POSCAR.
    """

    # Prequel: extract the list of chemical symbols from POSCAR in a convoluted way
    #-----------------------------------------------------------------------------
    with open(join(stordir,'POSCAR'),'r') as f: plines = f.readlines()
    uniqsymbs = plines[0].split() # ['H','O','Na']
    quantities = plines[5].split() # ['2','1','5']
    symbs = flatten([[a]*int(b) for a,b in zip(uniqsymbs,quantities)]) # ['H','H','O','Na','Na','Na','Na','Na']

    # Auxillary Functions
    #-------------------
    def l2vec(l:str)->list:
        """The first three numbers of a cell line contain x,y,z information"""
        return [roundfloat(x) for x in l.split()[:3]]

    def l2force(step:int,ind:int,l:str)->tuple:
        """
        The last three lines of a force line contain fx,fy,fz information
        A 'force' tuple needs the step number and atom number in its tuple, too
        """
        _,_,_,fx,fy,fz = l.split()
        return (step,ind,roundfloat(fx),roundfloat(fy) ,roundfloat(fz))

    def parse_v_section(sect    : str
                       ,stepnum : int
                       ) -> Tuple[Atoms
                                  ,int
                                  ,float
                                  ,List[ForceTuple]
                                  ,int]:
        """
        For a single ionic step, make a tuple with (Atoms,stepnum,energy,forces)
        """
        cell_text,cell_ind = btw(sect
                                 ,'reciprocal lattice vectors'
                                 ,'length of vectors')

        celllines = cell_text.split('\n')[1:-2]
        cell      = [l2vec(x) for x in celllines] # 3x3 matrix-like thing


        pf_text,pf_ind  = btw(sect
                             ,'POSITION'
                             ,'total drift'
                             ,cell_ind)

        lines       = pf_text.split('\n')[2:-2]
        posforces   = [(l2vec(x),l2force(stepnum,i,x)) for i,x in enumerate(lines)]
        pos,forces  = zip(*posforces)

        atoms  = Atoms(symbols   = symbs
                      ,positions = pos # <number of atoms> x 3 matrix thing
                      ,cell      = cell)

        atoms.wrap()
        eng = float(parse_line(sect[pf_ind:],'energy\(sigma->0\)',-1).split()[-1]) # "energy\(sigma->0\) = -73.2442"

        assert len(forces)==len(atoms), 'bad parsing of forces ... probably? (%d != %d)'%(len(forces),len(atoms))

        return (atoms,stepnum,eng,forces,0)

    # MAIN PROGRAM
    #-------------
    with open(join(stordir,'OUTCAR'),'r') as f: outcar = f.read()
    sections = outcar.split('EDIFF is reached')[1:]

    output = [parse_v_section(x,n) for n,x in enumerate(sections)]
    output[-1] = tuple([*output[-1][:-1],1]) # set final for finaltraj
    return output



from typing import List,Tuple
from networkx import MultiGraph,Graph  # type: ignore

def layers(MG:MultiGraph)->Tuple[List[int],List[int]]:
    """
    Assign constrained atoms as 'layer 0'.
    Atoms within one edge of layer n are labeled as layer n+1.
    """
    G = Graph(MG.copy()) # we don't care about multiple edges for this
    layer_dict = {d['index']:0 for _,d in G.nodes(data=True) if d['constrained']}
    curr_layer = 0
    if not layer_dict:
        return [],[]
    else:
        while len(layer_dict) < len(G):
            curr_atoms = [x for x in range(len(G)) if layer_dict.get(x) == curr_layer]
            for a in curr_atoms:
                for e in G.neighbors(a):
                    if e not in layer_dict:
                        layer_dict[e]=curr_layer+1
            curr_layer+=1
            if curr_layer > 50:
                for i in range(len(G)):
                    if i not in layer_dict:
                        layer_dict[i]=None
                break

        return list(layer_dict.keys()),list(layer_dict.values())

if __name__=='__main__':
    from dbgen.main                               import realDB
    from dbgen.support.utils                      import sqlselect
    from dbgen.scripts.Pure.Graph.json_to_graph   import json_to_graph
    import sys
    q = """SELECT geo_graph
            FROM alljob JOIN finaltraj USING (job_id)
                        JOIN struct USING (struct_id)
            WHERE  stordir = %s"""
    gstr = sqlselect(realDB,q,[sys.argv[1]])[0][0]
    G = json_to_graph(gstr)
    print(layers(G))



# External Modules
from abc import abstractmethod
from re import split
################################################################################

class SQLType(object):
    """
    SQL datatypes
    """
    @abstractmethod
    def __str__(self)->str:
        pass
    def __repr__(self)->str:
        return str(self)
    def __eq__(self,other:object)->bool:
        return self.__dict__==other.__dict__
    @staticmethod
    def from_str(s:str)->'SQLType':
        if 'VARCHAR' in s:
            mem = split(r'\(|\)',s)[1]
            return Varchar(int(mem))
        elif "DECIMAL" in s:
            prec,scale = split(r'\(|\)|,',s)[1:3]
            return Decimal(int(prec),int(scale))
        elif 'INT' in s:
            if   'TINY' in s: kind = 'tiny'
            elif 'BIG'  in s: kind = 'big'
            else:             kind = 'medium'
            signed = 'UNSIGNED' not in s
            return Int(kind,signed)
        elif 'TEXT' in s:
            if   'TINY' in s: kind = 'tiny'
            elif 'MED'  in s: kind = 'medium'
            elif 'LONG' in s: kind = 'long'
            else :            kind = ''
            return Text(kind)
        else:
            raise NotImplementedError("New SQLtype to parse?")
class Varchar(SQLType):
    """
    Varchar
    """
    def __init__(self,mem:int=255)->None:
        self.mem=mem
    def __str__(self)->str:
        return "VARCHAR(%d)"%self.mem

class Decimal(SQLType):
    def __init__(self,prec:int=10,scale:int=3)->None:
        self.prec  = prec
        self.scale = scale
    def __str__(self)->str:
        return "DECIMAL(%d,%d)"%(self.prec,self.scale)

class Int(SQLType):
    def __init__(self,kind:str='medium',signed:bool=True)->None:
        self.kind = kind
        self.signed = signed
    def __str__(self)->str:
        if   self.kind == 'tiny':   core= "TINYINT"
        elif self.kind == 'medium': core= "INTEGER"
        elif self.kind == 'big' :   core= "BIGINT"
        else:
            raise ValueError('unknown Int kind: '+self.kind)
        return core + "" if self.signed else " UNSIGNED"

class Text(SQLType):
    def __init__(self,kind:str='')->None:
        self.kind = kind
    def __str__(self)->str:
        if   self.kind == 'tiny':   return "TINYTEXT"
        elif self.kind == '':       return "TEXT"
        elif self.kind == 'medium': return "MEDIUMTEXT"
        elif self.kind == 'long' :  return "LONGTEXT"
        else:
            raise ValueError('unknown TEXT kind: '+self.kind)

class Date(SQLType):
    def __init__(self)->None:
        pass
    def __str__(self)->str:
        return "DATE"



from typing import Optional,List,Dict,Tuple
from json import loads,dumps
import numpy as np             # type: ignore
from numpy.linalg import solve # type: ignore

from dbgen.core.lists import flatten
from dbgen.core.numeric import common_integer

def find_reference_rxns(job     : int
                       ,job_ref : int
                       ,stoich  : str
                       ,refdata : Optional[str]
                       ) -> Tuple[str
                                    ,List[int]
                                    ,List[int]]:
    """
    Return a list of rxn names followed by unzipped <rxn,job,coef> triples
    We trust that the reference system will not have linearly dependent entries
    so we can just solve an Ax=b matrix problem where each row of A is a stoich
    vector of a reference, and b column is the stoich vector of the target job.

    If the input job is itself a reference, then we're in a tricky situation.
    We want to be able to query the reaction and find that its formation energy
    is zero, so there are two intuitive reactions to log: A (stoich 1) and A (stoich -1)
    or just A (stoich 0) If we did the first one, then that violates the PK
    constraint on job_id of stoich table, so we opt for the second strategy.

    """
    no_output = ('dummy',[],[]) # type: Tuple[str,List[int],List[int]]

    #####################
    # Auxillary Functions
    #--------------------
    def dict_to_vector(d     : Dict[int,int]
                      ,basis : List[int]
                      ) -> np.array:
        return np.array([d.get(b,0) for b in basis])

    def rxn_namer(out : List[Tuple[int,int]])->str:
        """This could definitely be improved to give coherent rxn names
          , though they'll still need to have job numbers since many different
          'rxns' have the same stoichiometry
          """
        return dumps(sorted(out))

    ##############
    # Main Program
    ##############

    # Process input strings
    #---------------------
    if refdata is None:
        return no_output
    elif job_ref > 0:
        return str(job),[job],[0]


    stoichdict = dict(loads(stoich))                     # {element : count}
    targetkeys = set(stoichdict.keys())
    allrefs    = [(a,b,dict(c)) for a,b,c in loads(refdata)] # [(JobID,{ElemID:Count})]
    refs       = [(a,b,c) for a,b,c in allrefs if not set(c.keys()).isdisjoint(targetkeys)]
    if refs == []:
        return no_output
    allelems   = list( targetkeys
                      | set(flatten([list(c.keys()) for _,_,c in refs])))

    # If multiple refs are for the same element, pick one at random
    # Later we'll try to do something better where multiple reactions are
    # output to accommodate all possibilities
    stoichmap = {x[1] : x for x in refs}

    uniqrefs  = list(stoichmap.values())

    # Set up matrices to solve for linear combination of references
    b = dict_to_vector(stoichdict,allelems)
    b = b.reshape(len(allelems),1)
    A = np.concatenate([dict_to_vector(r,allelems) for _,_,r in uniqrefs],axis=0)

    A.resize(len(allelems),len(allelems)) # make a square matrix, pad with zeros

    try:
        coef_array = solve(A,b)
    except:
        return no_output
    coefs = flatten(coef_array.tolist())


    output_ids = [job]+[uniqrefs[i][0] for i,coef in enumerate(coefs) if coef != 0]
    output_coefs = [-1.0]+coefs
    normed_coefs = common_integer(output_coefs)
    print('\t\tnormed %s to %s'%(output_coefs,normed_coefs))

    name = rxn_namer(list(zip(output_ids,normed_coefs)))
    print('named %s to %s'%(list(zip(output_ids,normed_coefs)),name))
    return name,output_ids,normed_coefs # type: ignore



from typing     import List,Tuple
from subprocess import getstatusoutput
from json       import loads
from os         import environ
################################################################################
def get_gpaw_logfiles(rootpath : str
                    ) -> Tuple[List[str],List[str]]:
    """
    Searches for DFT calculation folders. Returns unzipped (stordir,logfile) triples

        - If there exists a file in the directory with a particular signature,
            we identify it as a GPAW job.
    """
    ######################
    # Initialize Variables
    #---------------------
    dirs,logs  = [],[]
    badext     = ["sh","traj","json","gpw","py","err"]
    badextstr  = " ".join(['-not -name "*.%s"'%x for x in badext])
    user       = environ['SHERLOCK_USERNAME']

    ##############
    # Main program
    #-------------
    ssh     = 'ssh %s@login.sherlock.stanford.edu '%user

    cmd = """for f in $(find {0}  -type f {1})
            do
                if $(head -2 $f | tail -1 | grep -q " ___ ___ ___ _ _ _ "); then
                    if $(grep -q "Free energy" $f); then
                        echo $f
                    fi
                fi
        done""".format(rootpath,badextstr)
    exit,gpawcheck = getstatusoutput(cmd)
    assert exit==0,'Failure in get_gpaw_logfile bash execution: '+gpawcheck
    for logfile in gpawcheck.split('\n'):
        if logfile:
            last_slash = logfile.rfind('/')
            dir = logfile[:last_slash]
            if True: #dir not in files:
                dirs.append(dir)
                logs.append(logfile)

    return dirs,logs




def normalize_xc(xc : str) -> str:
    """
    Normalize the notation for xc
    """


    if xc.lower() == 'pbe':
        return 'PBE'
    elif xc.lower() in ['beef','beef-vdw']:
        return 'BEEF'
    elif xc.lower() == 'rpbe':
        return 'RPBE'
    elif xc.lower() == 'mbeef':
        return 'mBEEF'
    elif xc.lower() == 'hse':
        return 'HSE'
    elif xc.lower() == 'lda':
        return 'LDA'
    else:
        raise ValueError("weird xc: ",xc)



from typing import Any

import json
from dbgen.core.misc import cast_maybe

def get(dictstr: str
       ,key    : str
       ,dtype  : str
       ) -> Any:
    """
    Gets a key from a json'd dictionary
    """
    if not dictstr:  # catch cases of empty string and None
        return None

    assert dtype in cast_maybe.keys(), 'bad dtype argument for get: '+dtype

    try:
        d = json.loads(dictstr)
        out = d.get(key)

        return cast_maybe[dtype](out)
    except:
        print('failed to decode json in "GET"')
        import pdb;pdb.set_trace()



from typing import Dict

from dbgen.core.parsing import parse_line

def parse_xc_gpaw(log:str) -> str:
    """
    Finds the exchange correlation functional, bearing in mind the default
    """

    parsed =  parse_line(log,'xc:',0)
    if parsed is None:
        return 'LDA' # default
    else:
        return parsed.split(':')[-1].strip()



from typing import Tuple,List
from os     import environ
def populate_storage()->Tuple[List[str],List[str],List[str]]:
    """
    Reads DFT_LOGFILES to find a structured textfile with information
    about what directories should be considered 'roots'
    """
    with open(environ['DFT_LOGFILES'],'r') as f:
        rootdata  = [l.split() for l in f.readlines() if '#' not in l and l.strip()!='']
    paths,codes,labels = zip(*rootdata)
    return list(paths),list(codes),list(labels)



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_sigma_qe(log:str
             ) -> Optional[float]:
    """
    Extract Fermi-Dirac smearing from QE log file
    """
    # constants
    #----------
    ryd_to_ev = 13.60569

    # main program
    #------------
    parsed = parse_line(log,'Fermi-Dirac smearing',0)
    if parsed is None:
        raise ValueError('malformed QE logfile?')
    else:
        raw = parsed.split()[-1]
        return round(ryd_to_ev * float(raw),3)



from typing import Tuple
from dbgen.core.parsing import parse_line

def get_maxstep_gpaw(trip:Tuple[str,str,str]) -> int :
    """
    We need the dict instead of logfile because this info is only in pw.inp for QE
    """

    log,_,_ = trip
    parsed  = parse_line(log,'maxiter',0)
    if parsed is None:
        return 333 # default GPAW value
    else:
        return int(parsed.split(':')[-1])



from typing import Dict,Optional,Union
from dbgen.core.parsing import parse_line
from dbgen.core.misc    import cast_maybe

def get_vasp_key(dftcode : str
                ,outcar   : str
                ,key     : str
                ,dtype   : str
                ) -> Optional[Union[int,str,float]]:
    """
    common form of extracting VASP input info not present in QE/GPAW jobs

    If looking for a bool and key not found (e.g. LUSE_VDW), return False by default
    """

    assert dtype in cast_maybe.keys(), 'bad dtype argument for get_vasp_key: '+dtype

    if dftcode =='vasp':
        parsed = parse_line(outcar,key.upper(),0)
        if parsed is None:
            if dtype=='bool':
                return 0
            else:
                return None
        else:
            val = parsed.split('=')[-1].strip().split()[0]
            return cast_maybe[dtype](val) # type: ignore
    else:
        return None



from typing import List,Tuple,TYPE_CHECKING
if TYPE_CHECKING:
	from ase import Atoms # type: ignore

from pymatgen.io.ase 			import AseAtomsAdaptor 	  # type: ignore
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer # type: ignore
from string import ascii_lowercase
def get_spacegroup(atoms : 'Atoms')->int: #Tuple[int,List[int],List[int]]:
	"""
    uses pymatgen SpacegroupAnalyzer
	"""
	pmg = AseAtomsAdaptor().get_structure(atoms)
	sd  = SpacegroupAnalyzer(pmg,symprec=0.1).get_symmetry_dataset()
	sg   = sd['number']
	#wy   = [ascii_lowercase.index(x) for x in sd['wyckoffs']]

	return sg # (sg,list(range(len(atoms))),wy)



from typing import Tuple,TYPE_CHECKING
if TYPE_CHECKING:
    from bulk_enumerator import BULK # type: ignore
    
def get_pure_struct(b:'BULK')->Tuple[str,int,int]:
    """
    Extract info about a prototype structure
    """
    output =  (b.get_name()
              ,b.get_spacegroup()
              ,len(b.get_parameter_values()))
    b.delete() # deallocate memory
    return output



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_nmix_gpaw(log:str) -> Optional[float]:
    """
    Get the number of electronic steps mixed in for a new guess
    """
    parsed = parse_line(log,'nmaxold',0)
    if parsed is None:
        return 5 # default gpaw for systems with pbc
    else:
        raw    = parsed.split(':')[1].replace(',','')
        return float(raw)



from typing import Tuple,List

# External Modules
import json
from math      import gcd
from functools import reduce

# Internal Modules
from dbgen.core.lists import flatten

def load_comp(raw_structure : str
             ) -> Tuple[List[int],   List[int]
                           ,List[int],   List[int] , List[float]
                           ,List[float], List[int],  List[int]]:
    """
    Performs basic composition analysis, creating a row for every STRUCTURE x ELEMENT
    pair for which the element exists in the structure.

    Intended insert columns are:

              ['element_id',
               ,'count',	   	'count_fixed'
               ,'frac', 	   		'frac_fixed'
               ,'count_norm',  'count_fixed_norm']
    """

    # Load raw strings
    atomdata = json.loads(raw_structure)['atomdata']

    # Set up dictionaries of stoichiometries
    symbs     = [x['number'] for x in atomdata]
    setsymbs  = set(symbs)
    symbdict  = {s:symbs.count(s) for s in setsymbs}
    sum_      = sum(symbdict.values())
    denom     = reduce(gcd,symbdict.values(),0) # GCD of stoichiometry

    constsymbs = [x['number'] for x in atomdata if x['constrained']]
    fSymbdict = {s:constsymbs.count(s) for s in setsymbs}
    sumF = sum(fSymbdict.values())


    output = []
    for j in setsymbs:
        count       = symbdict[j]
        count_fixed = fSymbdict[j]

        frac        = round(count / sum_,3)
        frac_fixed  = 0 if sumF == 0 else round(count_fixed / sumF,3)

        count_norm = count/denom
        fixed_norm  = count_fixed/denom

        output.append((j,    count_fixed > 0
                      ,count,count_fixed,count_norm
                      ,frac, frac_fixed, fixed_norm))

    return list(map(list,zip(*output))) # type: ignore



from typing     import List,Tuple
from os         import environ,system
from os.path    import join,exists
from subprocess import getstatusoutput,Popen,PIPE
from json       import loads

from string     import ascii_uppercase,digits
from random     import choices
################################################################################
def get_catalog_logfiles(rootpath : str
                    ) -> Tuple[List[str],List[str],List[str],List[str],List[str],List[str],List[str],List[str],List[str]]:
    """
    Searches for DFT calculation folders. Returns unzipped (stordir,logfile,code) triples

    to get anytraj: \$(cat \$(ls \$f/*.traj | head -1))
    """
    ######################
    # Initialize Variables
    #---------------------
    user       = environ['SHERLOCK_USERNAME']
    login      = '%s@login.sherlock.stanford.edu'%user
    ssh        = 'ssh %s '%login
    delim1     = '***???***'
    delim2     = '~~~***~~~'
    rand       = ''.join(choices(ascii_uppercase+digits, k=16))

    # Main Program
    #-------------

    cmd    = r"""for f in \$(find {0} -name runtime.json -printf '%h\n')
                    do
                        if [ -f \$f/log ]; then
                            if \$(head -2 \$f/log | tail -1 | grep -q \" ___ ___ ___ _ _ _ \"); then
                                echo \"\$f{1}\$f/log{1}gpaw{1}\$(cat \$f/log){1} {1} {1} {1} \"
                            else
                                echo \"\$f{1}\$f/log{1}quantumespresso{1}\$(cat \$f/log) {1} \$(cat \$f/pw.inp) {1} {1} {1} \"
                            fi
                		else
                            if [ -f \$f/OUTCAR ]; then
                                echo \"\$f{1}\$f/OUTCAR{1}vasp{1}\$(cat \$f/OUTCAR){1} {1}\$(cat \$f/POTCAR){1}\$(cat \$f/POSCAR){1}\$(cat \$f/KPOINTS)\"
                            else
                                echo \"BAD DIRECTORY: \$f\"
                                exit 1
                            fi
                		fi
                        echo \" {1} \$(cat \$f/params.json) {2} \"
                   done""".format(rootpath,delim2,delim1)
    cmd1 = 'echo "%s" | %s "cat > %s; chmod 755 %s"'%(cmd,ssh,rand,rand)
    cmd3 = ssh + ' sh ' + rand
    cmd4 = ssh + ' rm ' + rand
    system(cmd1)
    code,output = getstatusoutput(cmd3)

    assert code == 0, output
    content = [[o.strip() for o in oo.split(delim2)]
                        for oo in output.split(delim1) if oo not in ['',' ']]

    try:
        dirs,logfiles,codes,logs,pws,pots,pos,kpts,params = zip(*content)
    except:
        import pdb;pdb.set_trace()
    system(cmd4)

    return (list(dirs),list(logfiles),list(codes),list(logs),list(pws)
           ,list(pots),list(pos),list(kpts),list(params))

if __name__ == '__main__':
    import sys
    print(get_catalog_logfiles(sys.argv[1]))



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_mixtype(dftcode : str
               ,log     : str
               ) -> Optional[str]:
    """
    Helpful docstring
    """

    if dftcode == 'quantumespresso':
        parsed = parse_line(log,'number of iterations used',0)
        if parsed is None:
            return None
        else:
            return parsed.split()[6]

    else:
        return None



from typing import List,Dict,Any,Optional,Tuple
# External Modules
from pprint import pprint,pformat
from decimal import Decimal

# Internal Modules
from dbgen.support.utils import sqlexecute,sqlselect,sqlexecutemany,mkInsCmd,sqlite_execute,ConnectInfo,addQs
from dbgen.support.datatypes.table import Table,FK
from dbgen.support.datatypes.misc import Arg,noIndex
################################################################################
################################################################################
class MetaTemp(object):
    """
    blocks   - define how the ruleed Query inputs are fed back into the DB
    """
    def __init__(self,blocks : List['Block']) -> None:
        self.blocks   = blocks

    def src(self)->str:
        return 'MetaTemp([%s])'%('\n\t,'.join(b.src() for b in self.blocks))

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__



    def rule_output_processor(self
                                  ,conn        : ConnectInfo
                                  ,result_dict : Dict[str,Any]
                                  ,verbose        : bool = False
                                  ) -> None:
        """
        Processes the output of a rule's query
        """

        for block in self.blocks:
            result_dict[block.name] = block.func(conn,result_dict,verbose)

    def add_metatemp(self
                    ,sqlite_pth   : str
                    ,rule_id : int
                    ) -> None:
        """
        Add to sqlite meta.db
        """
        for bId,block in enumerate(self.blocks):
            block.add_block(sqlite_pth,rule_id,bId)


class Block(object):
    """
    Execute a SQL statement after processing the Template.
        Inputs either refer to Template outputs ('stdout') or to other Blocks
        (by name, which should be unique)
        Possibly return something if statement is select.
    """
    def __init__(self
                ,name    : str
                ,text      : str
                ,args       : List[Arg] = []
                ) -> None:

        self.name    = name
        self.text    = text.strip()
        self.args    = args


    def src(self)->str:
        fmtargs = [self.name
                  ,'"""\n%s"""'%self.text
                  ,'\n\t\t\t,'.join(a.src() for a in self.args)]
        return "Block('{0}'\n\t\t,text={1}\n\t\t,args=[{2}])".format(*fmtargs)
    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def func(self
            ,conn        : ConnectInfo
            ,curr_dict : dict
            ,verbose   : bool = True
            ) -> Optional[List[tuple]]:
        """
        Creates a function which takes in a namespace dictionary and executes its
        block script (possibly using the namespace for inputs). Its result will
        get added to the namespace.
        """
        args = [arg.arg_get(curr_dict) for arg in self.args]
        maxlen = 1
        for a in args:
            assert isinstance(a,(int,str,float,list,bytes,Decimal,type(None))), "Arg (%s) BAD DATATYPE %s IN NAMESPACE "%(a,a.__class__)
            if isinstance(a,list):
                if maxlen != 1: # variable has been set
                    try:
                        assert(len(a) in [1,maxlen]), 'Can\'t broadcast: maxlen = %d, len a = %d (%s)'%(maxlen,len(a),str(a)) # preconditions for broadcasting
                    except:
                        import pdb;pdb.set_trace()
                else:
                    maxlen = len(a)

        def process_arg(x:Any)->list:
            if isinstance(x,list) and len(x)!=maxlen:
                return maxlen*x # broadcast
            elif not isinstance(x,list):
                return  maxlen * [x]
            else:
                return x

        # now all args should be lists of the same length
        broadcasted = [process_arg(x) for x in args]

        binds = list(zip(*broadcasted))

        if self.text[:6].lower()=='select':
            # we're probably selecting to get a PK, take the first result no matter what
            def app(x:tuple)->List[tuple]:
                    return sqlselect(conn,self.text,x)[0] # type: ignore

            def mbBox(x:tuple)->Any:
                if len(x)==1: return x[0]
                else: return x
            uniqbinds = list(set(binds))
            try:
                output = {b:mbBox(app(b)) for b in uniqbinds} # type: ignore
            except IndexError:
                    print('failed to find a result after query failed:\n'+self.text)
                    print('binds are : '+pformat(list(zip(self.args,uniqbinds[0]))))
                    import pdb;pdb.set_trace()
                    return []

            return [output.get(x) for x in binds] # type: ignore
        else:
            if len(self.args)==0:
                sqlexecute(conn,self.text)
            else:
                sqlexecutemany(conn,self.text,binds)
            return None

    def add_block(self
                 ,sqlite_pth : str
                 ,trans_id   : int
                 ,block_id   : int
                 ) -> None:
        """
        """
        block_icmd = mkInsCmd('block',['rule_id','block_id','text','name'],sqlite=True)
        sqlite_execute(sqlite_pth,block_icmd,[trans_id,block_id,self.text,self.name])
        for i,a in enumerate(self.args):
            a.add_arg(sqlite_pth,'meta',trans_id,block_id,i)

    @staticmethod
    def maybe_index(d   : dict
                   ,arg : Tuple[str,Optional[int]]
                   ) -> Any:
        """
        This is exact copy from (some version of) temp.py
        """
        if arg[1] is None:
            return d[arg[0]]
        else:
            return d[arg[0]][arg[1]]

################################################################################
################################################################################
# MetaTemp Shortcuts
#-------------------
def PureSQL(strs : List[str])->MetaTemp:
    """ Executes a series of SQL statements in order """
    return MetaTemp([Block(str(i),s) for i,s in enumerate(strs)])

def SimpleInsert(tab   : str
                ,cols  : List[str]
                ) -> MetaTemp:
    """
    Assume that the insert columns have corresponding variables in the namespace
    """
    args   = noIndex(cols)
    qmarks = ','.join(['%s']*len(cols))
    return MetaTemp([Block('insert'
                          ,"""INSERT INTO {0} ({1})
                                  VALUES ({2})
                                  ON DUPLICATE KEY
                                UPDATE {3} = {3}""".format(tab
                                                          ,','.join(cols)
                                                          ,qmarks
                                                          ,cols[0])
                          ,args)])

def SimpleUpdate(tab  : Table
                ,cols : List[str]
                ) -> MetaTemp:
    """
    Assume that the insert columns AND primary keys have corresponding variables
    in the namespace
    """
    pk_cols = [c.name for c in tab.cols if c.pk == True]
    args    = noIndex(cols+pk_cols)
    set     = addQs(cols,',')
    where   = addQs(pk_cols,' AND ')

    return MetaTemp([Block('update'
                          ,'UPDATE %s SET %s WHERE %s'%(tab.name,set,where)
                           ,args)])

def SimpleUpsert(ins_tab   : Table
                ,ins_cols  : List[str]
                ,fk_tab    : Table
                ) -> MetaTemp:
    """
    Let's say we compute an entity (A) associated with another entity (B).
    - We want to insert the A instance into the A table (if it doesn't exist).
    - We want to record the PK of this A instance an store it as a FK in B table.

    ins_tab   - 'A' table described above
    cols       - properties with which to insert an A instance
    fk_tab    - name of the 'B' table described above

    We assume that:
        - there exists exactly one foreign key from B to A
        - we are going to populate all columns of foreign key if it's composite
        - whatever was inserted will uniquely identify an A object (e.g. 'raw'
            for structure, possibly not necessarily the PKs of A)
        - variables from template associated with various columns are equal to
          their column names

    WARNING THIS WILL BE A PROBLEM IF THERE'S A NAMESPACE COLLISION IN KEY NAMES

    SAFER WOULD BE TO MAKE THE 'key' FOR COLUMNS TO BE '<table>.<column>'
    """

    insert_args   = noIndex(ins_cols)
    insert_qmarks = ','.join(['%s']*len(ins_cols))
    fmt_args0     = [ins_tab.name,','.join(ins_cols),insert_qmarks,ins_cols[0]]
    insert_block  = Block('insert'
                         ,"""INSERT INTO {0} ({1})
                             VALUES ({2}) ON DUPLICATE KEY
                            UPDATE {3}={3}""".format(*fmt_args0)
                         ,insert_args)

    ins_pks = [c.name for c in ins_tab.cols if c.pk]
    where  = addQs(ins_cols,' AND ')
    fmt_args1 = [','.join(ins_pks),ins_tab.name,where]
    get_id_block = Block('get_id'
                        ,'SELECT {0} FROM {1} WHERE {2}'.format(*fmt_args1)
                        ,insert_args)

    fk              = fk_tab.get_fk(ins_tab)
    fk_pks          = [c.name for c in fk_tab.cols if c.pk]
    setvals      = addQs(fk.column,' , ')
    whereagain   = addQs(fk_pks,' AND ')
    args          = [Arg('get_id',i) for i in range(len(fk.column))]
    fmt_args2    = [fk_tab.name,setvals,whereagain]
    args.extend(noIndex(fk_pks))

    update_block = Block('update'
                        ,'UPDATE {0} SET {1} WHERE {2}'.format(*fmt_args2)
                        ,args)

    return MetaTemp([insert_block, get_id_block, update_block])



from typing import Optional,List,Dict,Any,Callable

# External Modules
from time            import time
from pprint          import pformat
from tqdm            import tqdm                   # type: ignore
from multiprocessing import Pool,cpu_count
from sqlparse        import format as sql_format   # type: ignore

# Internal Modules
from dbgen.support.sqlparsing         import parse_dependencies
from dbgen.support.datatypes.metatemp import MetaTemp
from dbgen.support.datatypes.temp     import Temp
from dbgen.support.datatypes.func     import Func
from dbgen.support.datatypes.misc     import ExternalError
from dbgen.support.utils              import (sqlexecute,sqlselect,mkInsCmd,sqlite_execute
                                             ,namespaceCmd,func_to_ind,sqlite_select
                                             ,handle_to_ind,ConnectInfo,select_dict)
from dbgen.core.lists   import flatten,concat_map
from dbgen.core.parsing import btw
from dbgen.core.numeric import safe_div
#########################################################################################

class Rule(object):
    """
    Information required to transform the database

    Generally, one must provide:
        - a unique name by which this action can be referred (method_name)
        - the name of the entity/table that is being altered (tab_name)
        - whether or not we're inserting new rows vs updating columns (insert)
        - what prerequiste information is needed beforehand (deps)
        - what information we will be providing (cols)
        - how to get the required information for the rule (query)
        - what operation to transform the inputs (func)
    """

    def __init__(self
                ,name       : str
                ,query      : Optional[str] = None
                ,metatemp   : MetaTemp      = MetaTemp([])
                ,template   : Temp          = Temp([],{})
                ) -> None:

        if query is not None:
            query = sql_format(query, reindent=True).strip()
            query = sql_format(query, reindent=True) # indempotency

        self.name       = name
        self.metatemp   = metatemp
        self.query      = query
        self.template   = template

        # Precompute dependencies of all SQL commands
        mbQuery = [] if self.query is None else [self.query]
        allsql = mbQuery + [b.text for b in self.metatemp.blocks]
        self.deps   = parse_dependencies(allsql)

    def src(self)->str:
        query = '' if self.query is None else '\n\t\t,query="""\n%s"""'%self.query
        temp  = '' if self.template == Temp([],{}) else "\n\t,template=%s"%self.template.src()
        fmtargs = [self.name,query,temp,self.metatemp.src()]
        return "Rule('{0}'{1}{2}\n\t,metatemp={3})".format(*fmtargs)

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__


    def dependency_failed(self
                         ,conn : ConnectInfo
                         ) -> Any:
        q = """SELECT child
                FROM META_dependencies
                JOIN META_rule ON parent = method
                WHERE child =%s AND status like '%%failed%%'"""
        output = sqlselect(conn,q,[self.name])
        return True if len(output) == 0 else output

    def run_rule(self
                ,conn    : ConnectInfo
                ,verbose : bool = True
                ,parallel: bool = True
                ) -> None:
        """
        Executes a SQL query, then maps each output over a processing function
        which modifies the database.
        """

        try:
            start = time()

            if self.query is None:
                inputs = [{}] # type: List[dict]
            else:
                inputs = list(select_dict(conn,self.query)) # memory problems if we convert iterator to list?

            q = "UPDATE META_rule SET n_inputs=%s WHERE method=%s"
            sqlexecute(conn,q,[len(inputs),self.name])

            process_raw_args = self.zip_many((self,conn),inputs)
            if parallel:
                cpus = cpu_count()-1 or 1 # play safe, leave one free
                with Pool(cpus + 1 ) as p: # catastrophic failure less likely if we use n-1 available nodes
                    for _ in tqdm(p.imap_unordered(self.process_row,process_raw_args)
                                 ,total=len(inputs)):
                        pass
            else:
                for row in tqdm(process_raw_args):
                    self.process_row(row)

            self.update_status(conn,'completed')
            tot_time = time()-start
            self.update_time(conn,tot_time/60,safe_div(tot_time,len(inputs)))

        except ExternalError as e:
            err = str(e).replace('\\n','\n')
            msg = 'Error when running rule %s\n'%self.name
            print(msg + err)
            self.log_error(conn,msg + err)

    def log_error(self
                 ,conn : ConnectInfo
                 ,err  : str
                 ) -> None:
        self.update_status(conn,'failed')
        q = "UPDATE META_rule SET error=%s WHERE method=%s"
        sqlexecute(conn,q,[err,self.name])

    def update_status(self
                     ,conn : ConnectInfo
                     ,stat : str
                     ) -> None:
        q = "UPDATE META_rule SET status=%s WHERE method=%s"
        sqlexecute(conn,q,[stat,self.name])

    def update_time(self
                   ,conn     : ConnectInfo
                   ,duration : float
                   ,rate     : float
                   ) -> None:
        q = "UPDATE META_rule SET runtime=%s,rate=%s WHERE method=%s"
        sqlexecute(conn,q,[duration,rate,self.name])

    def add_rule(self,sqlite_pth : str) -> None:
        """
        Insert a Rule to the meta.db
        """
        cmd   = mkInsCmd('rule',['name','query'],sqlite=True)
        sqlite_execute(sqlite_pth,cmd,[self.name,self.query])
        sqlite_execute(sqlite_pth,namespaceCmd,[self.name,'Rule'])
        t_id = handle_to_ind(sqlite_pth,self.name)
        self.metatemp.add_metatemp(sqlite_pth,t_id)
        self.template.add_temp(sqlite_pth,t_id)

    @staticmethod
    def zip_many(t:tuple,xs:list)->List[tuple]:
        return [(*t,x) for x in xs]

    @staticmethod
    def process_row(tup : tuple) -> None:
        (t,conn,row) = tup          # unpack
        try:
            t.metatemp.rule_output_processor(conn
                ,t.template.apply_template(row))
        except ExternalError as e:
            def abbreviate(x:Any)->Any:
                if isinstance(x,str) and len(x) > 1000:
                    return x[:1000]+'...'
                else:
                    return x
            row_ = {k:abbreviate(v) for k,v in row.items()}
            raise ExternalError('%s failed with inputs %s \n%s'%(t.name,row_,e))



from typing import Dict,Optional
import json
from dbgen.core.parsing import parse_line

def get_dipol_vasp(outcar : str) -> Optional[str]:
    """
    Docstring
    """
    line   = parse_line(outcar,' DIPOL')
    if line is None:
        return None
    else:
        print('parsing dipol line: '+line)
        raw    = line.split('=')[-1].split()
        output = [float(x) for x in raw]
        return json.dumps(output)



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_sigma_vasp(outcar:str) -> Optional[float]:
    """
    Helpful docstring
    """

    parsed = parse_line(outcar,'SIGMA',0)
    if parsed is None:
        raise ValueError('malformed OUTCAR? no "SIGMA" found')
    else:
        raw    = parsed.split('=')[1].strip().split()[0]
        return float(raw)



from typing import Dict
from dbgen.core.parsing import parse_line

def normalize_psp(pspth : str) -> str:
    """
    Normalizes pseudopotential name
    """

    if   'gbrv'   in pspth: return 'gbrv15pbe'
    elif 'sg15'   in pspth: return 'sg15'
    elif 'paw'    in pspth:  return 'paw'
    elif pspth[-6:] == 'LDA.gz':  return 'lda_paw'
    elif 'oldpaw'   == pspth:     return 'oldpaw'
    elif 'setups-0.8.7929' in pspth: return 'gpaw-setups-0.8.7929'
    elif '{}'       == pspth:     return '{}'
    elif 'dacapo' in pspth or 'esp-psp' in pspth or 'pslib' in pspth: return 'dacapo'
    else:
        raise NotImplementedError('New psp? path = ',pspth)



from typing import Tuple
from dbgen.core.parsing import parse_line
################################################################################
def get_maxstep_vasp(trip:Tuple[str,str,str]) -> int :
    """
    We need the dict instead of logfile because this info is only in pw.inp for QE
    """

    outcar,_,_ = trip
    parsed = parse_line(outcar,'NELM',0)
    if parsed is None:
        raise ValueError('malformed outcar?')
    else:
        raw = parsed.split('=')[1].strip().split()
        return int(raw[0].replace(';',''))



from typing import Dict
from dbgen.core.parsing import parse_line

def parse_psp_gpaw(log:str) -> str:
    """
    Returns simplified string representation of the pseudopotnetial used in a
    DFT calculation
    """
    rawpspth = parse_line(log,'setups',0)

    if rawpspth is None:
        return 'paw' # default
    else:
        pspth = rawpspth.split(': ')[-1].strip()
        if 'gpaw-setups-0.6.6300' in pspth:
            return 'oldpaw'
        elif 'setups/' == pspth[:7]:
            return 'paw'
        else:
            return pspth



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ase import Atoms # type: ignore
from ase.io import write,read # type: ignore
from bulk_enumerator.bulk import BULK # type: ignore
from os import remove
from string import ascii_uppercase,digits
from random import choices

def get_bulk(atoms : 'Atoms'
            ,tol   : float  = 0.05
            ) -> BULK:
    """
    Create a BULK object from Ankit's library
    """
    # Constants
    #----------
    pth = ''.join(choices(ascii_uppercase+digits, k=16))+'_POSCAR'
    # Initialize
    #-----------
    assert tol <= 0.1 # requirement?
    b = BULK(tolerance=tol)
    poscar = ''
    # Get POSCAR String
    #------------------
    while poscar == '':
        write(pth,atoms,format='vasp')
        try:
            with open(pth,'r') as f:
                poscar = f.read()
        except IOError:
            pass # try repeatedly

    # Create bulk from POSCAR string
    #-------------------------------
    b.set_structure_from_file(poscar)

    if b.get_name()=='':
        raise ValueError('problem with getting bulk: \n'+poscar)
    try:
        remove(pth)
    except IOError:
        pass

    return b



from typing import Tuple

def cell_info(ax:float
             ,ay:float
             ,az:float
             ,bx:float
             ,by:float
             ,bz:float
             ,cx:float
             ,cy:float
             ,cz:float
             ) -> Tuple[float,float,float,float,float]:
    """
    Details about a cell that we compute from its 3 defining vectors
    """
    surf_area = ((ax*bz-az*by) * (ay*bz - az*by)
                +(ax*bz-az*bx) * (ax*bz - az*bx)
                +(ax*by-ay*bx) * (ax*by - ay*bx)) ** 0.5

    vol = ( ax * (by * cz - bz * cy)
          - ay * (bx * cz - bz * cx)
          + az * (bx * cz - bz * cx))

    a   = (ax**2 + ay**2 + az**2)**0.5
    b   = (bx**2 + by**2 + bz**2)**0.5
    c   = (cx**2 + cy**2 + cz**2)**0.5

    return surf_area,vol,a,b,c



from typing import List,Tuple,TYPE_CHECKING
if TYPE_CHECKING:
    from ase import Atoms # type: ignore
from pymatgen.io.ase 			import AseAtomsAdaptor 	         # type: ignore
from pymatgen.analysis.structure_analyzer import VoronoiAnalyzer # type: ignore

def get_voronoi(atoms:'Atoms')->Tuple[List[int],List[int],List[int],List[int]
                                 ,List[int],List[int],List[int],List[int]
                                 ,List[int]]:
    """
    Voronoi analysis returns # of 3-10 sided polyhedra. Error yields -1 values.
    """
    pmg = AseAtomsAdaptor().get_structure(atoms)
    inds = list(range(len(atoms)))
    try:
        out = [[i,*VoronoiAnalyzer().analyze(pmg,i)] for i in inds]
        return tuple(map(list,zip(*out)))  # type: ignore
    except:
        return tuple([inds] + [[-1]*len(atoms)]*8)   # type: ignore



from typing import Dict
from dbgen.core.parsing import parse_line

def parse_psp_vasp(outcar   : str) -> str:
    """
    Don't know how to do this yet...
    ideally we parse the potcar to reconstruct setups dictionary
    """
    return '{}'



import json

def mk_relaxed_surface(rawstrs   : str
                      ,adsstrs   : str
                      ,orderstr  : str
                      ,stordirs  : str
                      ,facetstrs : str
                      ) -> None:
    """
    Useful object for calculation intermediate
    # """
    raise NotImplementedError
    # raw1,raw2 = list(map(json.loads,rawstrs.split('@@@')))
    # ads1,ads2 = adsstrs.split('@@@')[0],adsstrs.split('@@@')[1]
    # orderstrs = orderstr.split('@@@')
    # sd1,sd2   = stordirs.split('@@@')
    # fa1,fa2   = list(map(json.loads,facetstrs.split('@@@')))
    #
    # class RelaxedSurface(object):
    #     """columns needed: ."""
    #     def __init__(self,raw1,raw2,ads1,ads2,sd1,sd2,fa1,fa2): # type: ignore
    #
    #         self.initatoms  = json_to_traj(raw1) # ase.Atoms
    #         self.finalatoms = json_to_traj(raw2) # ase.Atoms
    #         self.adsorbates = json.loads(adsorbates)                          # [[Int]]
    #         self.facet      = json.loads(facet)                               # (Int,Int,Int)
    #         self.cell       = json.loads(cell)
    #         self.bare       = self.initatoms.copy()                           # ase.Atoms
    #         #self.stordir    = storage_directory                               # for debugging purposes
    #         for i in flatten(self.adsorbates): del self.bare[i]
    #
    #
    #     def coverage(self):  return len(self.adsorbates)/float(len(self.sites)) # type: ignore
    #
    #     def get_adsorbates(self):  # type: ignore
    #         """ Get a list of ase.Atoms objects, corresponding to gas phase
    #             representations of the adsorbates, if any"""
    #         output = []
    #         for ads in self.adsorbates:
    #             ads_atoms = self.finalatoms.copy()
    #             for i in ads: del ads_atoms[i]
    #             output.append(ads_atoms.copy())
    #         return output
    #
    #     def _get_sites(self,site_type='all',sym_reduce=0): # type: ignore
    #         """Use pymatgen site finder to locate surface sets"""
    #         return [ads_site.Site(self.bare,s,self.facet) for s in get_sites(self.bare,self.facet,site_type,sym_reduce)]
    #
    #     def get_xy(self): # type: ignore
    #         """Approximately get x,y dimensions (in # of atoms) of a slab """
    #         [lx,ly,_] = map(np.linalg.norm,self.cell)
    #         xyproduct = len(self._get_sites(site_type='ontop')) # x * y = a
    #         ratio     = lx / ly                                 # x / y = b
    #
    #         def rounder(z): return int(max(1,round(z)))  # type: ignore
    #         x = rounder((xyproduct * ratio)**0.5)  # type: ignore
    #         y = rounder(x / ratio)                  # type: ignore
    #         if x*y == xyproduct: return (x,y)
    #         else:
    #             if 2 * x * y == xyproduct: return (2*x,2*y)
    #             else:
    #                 print('something wrong in get_xy: xyproduct=%d,ratio=%f,x=%d,y=%d \n%s'%(xyproduct,ratio,x,y,self.stordir))
    #                 return (None,None)
    #
    #     def get_layers(self): # type: ignore
    #         """Approximately get the number of layers of a slab"""
    #         zs    = [p[2] for p in self.bare.get_positions()]
    #         zlist = [zs[0]]
    #         for z in zs:
    #             if all([abs(zz -z) > 1.5 for zz in zlist]): zlist.append(z)
    #         return len(zlist)
    #
    #     def assign_sites(self): # type: ignore
    #         """Maps list of adsorbates to list of Site objects"""
    #         ads_sites = []
    #         for ads in self.get_adsorbates():
    #             com = ads.get_center_of_mass()
    #             _,min_site = min([(np.linalg.norm(com-s.pos),s) for s in self.get_sites()])
    #             ads_sites.append(min_site)
    #         return ads_sites
    #
    #     def reconstructed(self): # type: ignore
    #         """True if surface reconstructed"""
    #         raise NotImplementedError
    #
    #     def surface_area(self): # type: ignore
    #         return np.linalg.norm(np.cross(self.cell[0],self.cell[1]))
    #
    #     def constraints_in_middle(self): # type: ignore
    #         """
    #         is it a 'symmetric' slab or not?
    #         just testing if constrained atoms are not at bottom
    #         use case: we don't care about 'surface energy' unless slab is symmetric
    #         """
    #         atoms = self.finalatoms.copy()
    #         atoms.center()
    #         minConstrained,minUnconstrained = 100000,100000
    #         for a in atoms:
    #             if a.index in atoms.constraints[0].get_indices():
    #                 if a.z < minConstrained: minConstrained = a.z
    #             elif a.z < minUnconstrained: minUnconstrained=a.z
    #         return minUnconstrained < minConstrained
    #
    #     def get_vacuum(self): # type: ignore
    #         """Length of vacuum layer between periodic images (in Z direction)"""
    #         atoms = self.bare.copy()
    #         atoms.center()   # In case dealing with Nerds Rope & Co.
    #         c     = np.linalg.norm(atoms.get_cell()[2])
    #         zs    = [a.position[2] for a in atoms]
    #         return c - (max(zs) - min(zs))
    #
    # return RelaxedSurface(raw1,raw2,ads1,ads2,sd1,sd2,fa1,fa2)



from typing     import List,Tuple
from os         import environ
from os.path    import join,exists
from subprocess import getstatusoutput,Popen,PIPE
from json       import loads
################################################################################
def get_dft_logfiles(rootpath : str
                    ) -> Tuple[List[str],List[str],List[str]]:
    """
    Searches for DFT calculation folders. Returns unzipped (stordir,logfile,code) triples

    Requirements to be flagged as a DFT job:
    """
    ############################
    # Initialize Variables
    #---------------------
    out        = []
    user       = environ['SHERLOCK_USERNAME']
    # Main Program
    #-------------
    ssh    = 'ssh %s@login.sherlock.stanford.edu '%user
    cmd    = ssh+"find {} -name runtime.json ".format(rootpath)
    output = getstatusoutput(cmd)[1]
    for logfile in output.split('\n'):
        print('logfile = '+logfile)
        if logfile:
            last_slash = logfile.rfind('/')
            root       = logfile[:last_slash]
            files      = getstatusoutput(ssh+"ls %s"%root)[1].split()
            if 'log' in files:
                logpth   = join(root,'log')
                log      = getstatusoutput(ssh+'cat %s'%logpth)[1]
                is_gpaw  = ' ___ ___ ___ _ _ _ ' in log[:500]
                code     = 'gpaw' if is_gpaw else 'quantumespresso'
                out.append((root,logpth,code))
            elif 'OUTCAR' in files:
                out.append((root,join(root,'OUTCAR'),'vasp'))
            else:
                errstr = 'Malformed CataLog directory? No log nor OUTCAR found \n'
                raise ValueError(errstr+root)

    dirs,logs,codes = zip(*out)
    return (list(dirs),list(logs),list(codes))

if __name__ == '__main__':
    import sys
    print(get_dft_logfiles(sys.argv[1]))



from typing import Dict,Tuple

def get_kpts_qe(trip:Tuple[str,str,str]) -> Tuple[int,int,int]:
    """
    K point info (? ? ? kx ky kz) assumed to be in penultimate line of pw.inp
    """

    _,pwinp,_ = trip
    line  = pwinp.split('\n')[-1]
    raw   = line.split()[:3]
    return tuple([int(x) for x in raw])    #type: ignore



from typing import Dict

from dbgen.core.parsing import parse_line

def parse_dftcode(logfile : str) -> str:
    """
    gets a dftcode, given the logfile identified in get_dft_logfiles
    """
    if 'POTCAR' in logfile[:2000]:
        return 'vasp'
    elif '___ ___ ___ _ _ _' in logfile[:500]:
            return 'gpaw'
    elif 'Quantum ESPRESSO' in logfile[:2000]:
        return 'quantumespresso'
    else:
        raise ValueError('could not determine DFT code ')



from typing import TypeVar,Callable,List
A = TypeVar('A')
B = TypeVar('B')

def map_(f    : Callable[[A],B]
        ,args : List[A]
        ) -> List[B]:
    """
    Maps a function over an input. If each input has multiple outputs, than
    multiple lists are generated.
    """
    mapped = [f(arg) for arg in args]

    if mapped == []:
        return []
    else:
        if not isinstance(mapped[0],tuple):
            return mapped
        else:
            return tuple(map(list,zip(*mapped)))



from typing import List,Dict
from  ase  import Atoms                         # type: ignore
from ase.constraints import FixAtoms            # type: ignore
import networkx as nx                           # type: ignore
import numpy as np                              # type: ignore
from scipy.cluster.hierarchy import fclusterdata #type: ignore
from json import loads
from collections import defaultdict

def mk_graph(atoms        : 'Atoms'
            ,include_frac : float
            ,group_cut    : float
            ,min_bo       : float
            ,bonddata     : str
            ,bothways     : bool       = False
            ) -> nx.MultiGraph:
    """
    Create a graph
    Bonddata is a json'd list of tuples containing the following Edge information
        (fromNode  : int
        ,toNode    : int
        ,bondorder : float
        ,x         : int
        ,y         : int
        ,z         : int)
    """

    # Auxillary Functions
    #--------------------
    def make_adjacency_matrix(es : List[dict])->List[List[float]]:
        """
        Square matrix where (i,j) is the sum of bond order between atoms i and j
        """
        output = np.zeros((len(atoms),len(atoms)))   # type: ignore
        for edge in es:
            output[edge['fromNode'],edge['toNode']] += edge['bondorder']
        return output.tolist()

    def make_group(edge_list:List[dict])->List[List[dict]]:
        """
        Partitions a set of Edges into groups with similar bond strength
        """

        # Handle edge cases
        #==================
        if len(edge_list)==0:  return []
        if len(edge_list)==1:  return [edge_list]

        strs = np.array([[e['bondorder'] for e in edge_list]]).T # create (n_edges,1) array
        groups = defaultdict(list)  #type: dict

        group_inds = list(fclusterdata(strs,group_cut
                                ,criterion='distance',method='ward'))

        for i in range(len(edge_list)): groups[group_inds[i]].append(edge_list[i])
        maxbo_groups = [(max([e['bondorder'] for e in es]),es) for es in groups.values()]
        sorted_maxbo_groups = list(reversed(sorted(maxbo_groups)))
        return [es for maxbo,es in sorted_maxbo_groups]

    def get_filtered_edges(bonds : List[dict])->List[dict]:
        if bonds == []: return [] # (no) edge case
        output = [] # type: List[dict]
        edge_dict = defaultdict(list) # type: Dict[int,List[dict]]
        for bond in bonds:
            edge_dict[bond['fromNode']].append(bond)

        total_bo = {ind:sum([e['bondorder'] for e in es])
                            for ind,es in edge_dict.items()}
        groups   = {ind: make_group(es)
                            for ind,es in edge_dict.items()}

        max_ind  = max([max(e['fromNode'], e['toNode']) for e in bonds])

        for ind in groups.keys():
            if total_bo[ind] > 0:
                accumulated = 0.0
                for i, group in enumerate(groups[ind]):
                    output.extend(group)
                    accumulated += sum([e['bondorder'] for e in group])/total_bo[ind]
                    if accumulated > include_frac:
                        break
        return output

    # Main Program
    #--------------
    attrs = ['toNode','fromNode','bondorder','x','y','z']
    if bonddata == '[]':
        bonds = [] # type: List[dict]
    else:
        bonds = [dict(zip(attrs,args)) for args in loads(bonddata)]
        if bothways:
            for b in bonds:
                if b['fromNode'] <= b['fromNode']:
                    bonds.append(dict(zip(attrs,(b['toNode']
                                     ,b['fromNode']
                                     ,b['bondorder']
                                     ,-b['x'],-b['y'],-b['z']))))

    G  = nx.MultiGraph(cell = atoms.get_cell().tolist())  # Initialize graph

    fixed_inds = []
    if atoms.constraints:
        for constraint in atoms.constraints:
            if isinstance(constraint,FixAtoms):
                fixed_inds.extend(list(constraint.get_indices()))

    for i in range(len(atoms)): # add nodes
        G.add_node(i,symbol      = atoms[i].symbol
                    ,number      = int(atoms[i].number)
                    ,position    = atoms[i].position.tolist()
                    ,index       = i
                    ,constrained = int(i in fixed_inds)
                    ,magmom      = atoms[i].magmom)

        edges = get_filtered_edges(bonds)

    for e in edges:
        prop_dict = {'bondorder':e['bondorder']
                    ,'pbc_shift': (e['x'],e['y'],e['z'])}
        tupl = (e['toNode'],e['fromNode'])
        G.add_edges_from([tupl],**prop_dict)

    adj_matrix = make_adjacency_matrix(edges)

    G.graph['adj_matrix'] = adj_matrix

    return G



from typing import Optional,Any,List,TYPE_CHECKING
if TYPE_CHECKING:
    from dbgen.support.datatypes.object   import Object
from copy import deepcopy

from dbgen.support.datatypes.sqltypes   import SQLType,Int
from dbgen.support.datatypes.table      import Col
from dbgen.support.datatypes.constraint import EQ,NE,LT,GT
###############################################################################
class Attr():
    """
    A property of an object - needs a name and datatype
    Can be given a default value and NOT NULL constraint
    """
    def __init__(self
                ,name    : str
                ,dtype   : SQLType       = Int()
                ,nnull   : bool          = False
                ,ind     : bool          = False
                ,default : Optional[Any] = None
                ,uniq    : bool          = False
                ,auto    : bool          = False
                ,pk      : bool          = False
                ,obj     : str           = '???'
                ) -> None :

        self.name    = name
        self.dtype   = dtype
        self.nnull   = nnull
        self.ind     = ind
        self.default = default
        self.uniq    = uniq
        self.auto    = auto
        self.pk      = pk
        self.obj     = obj

    # Basic
    #------

    def __str__(self) -> str:
        return '%s.%s'%(self.obj,self.name)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self,other : object)->EQ: # type: ignore
        return EQ(self,other)

    def __ne__(self,other : object)->NE: # type: ignore
        return NE(self,other)

    def __lt__(self,other : object)->LT: # type: ignore
        return LT(self,other)

    def __gt__(self,other : object)->GT: # type: ignore
        return GT(self,other)

    def __hash__(self)->int:
        return hash(self.name)

    # Other methods
    #--------------
    def _mkCol(self)->Col:
        return Col(name    = self.name
                  ,typ     = self.dtype
                  ,nn      = self.nnull
                  ,uniq    = self.uniq
                  ,ind     = self.ind
                  ,default = self.default
                  ,auto    = self.auto
                  ,pk      = self.pk)



#External Modules
from typing import List,TYPE_CHECKING
if TYPE_CHECKING:
    from dbgen.support.datatypes.object     import Schema

from abc    import abstractmethod
#Internal Modules
from dbgen.support.datatypes.attr       import Attr
from dbgen.support.datatypes.metatemp   import Block
from dbgen.support.datatypes.misc       import noIndex
from dbgen.support.utils                import addQs
################################################################################

class InsertUpdate(object):
    @abstractmethod
    def mk_block(self)->Block:
        raise NotImplementedError

class Insert(InsertUpdate):
    """
    Take items in a namespace and store them as attributes on objects
    in the database
    """
    def __init__(self
                ,schema : 'Schema'
                ,attrs  : List[Attr]
                ) -> None:
        self.obj = schema.get_(attrs[0].obj)
        assert all([a.obj == self.obj._name for a in attrs])
        self.attrs = attrs
        self.n     = len(attrs)

    def mk_block(self)->Block:
        colnames = [a.name for a in self.attrs]
        qs       = ','.join(['%s']*self.n)
        name     = 'insert_%s_%s'%(self.obj._name,','.join(colnames))
        text     = 'INSERT INTO %s (%s) VALUES (%s)'%(self.obj._name
                                                   ,','.join(colnames)
                                                   ,qs)
        dup      = ' ON DUPLICATE KEY UPDATE {0} = {0}'.format(colnames[0])
        args = noIndex(colnames)
        return Block(name = name
                    ,text = text + dup
                    ,args = args)


class Update(InsertUpdate):
    """
    Update attributes of an object
    """
    def __init__(self
                ,schema : 'Schema'
                ,attrs : List[Attr]
                ) -> None:
        self.schema = schema
        self.obj = schema.get_(attrs[0].obj)
        assert all([a.obj == self.obj._name for a in attrs])
        self.attrs = attrs

    def mk_block(self)->Block:
        colnames = [a.name for a in self.attrs]
        pk_cols  = [a.name for a in self.obj._pks(self.schema)]
        name     = 'update_%s_%s'%(self.obj._name,','.join(colnames))
        setstr   = addQs(colnames,',')
        where    = addQs(pk_cols,' AND ')
        text     = 'UPDATE %s SET %s WHERE %s'%(self.obj._name,setstr,where)
        args = noIndex(colnames+pk_cols)
        return Block(name = name
                    ,text = text
                    ,args = args)



from typing import Dict,Optional

from dbgen.core.parsing import parse_line

def get_dw(dftcode : str
          ,log : str
          ) -> Optional[int]:
    """
    Density wave cutoff, only meaningful for quantumespresso
    """
    # Constants
    #-----------
    ryd_to_ev =  13.60569

    # Main Program
    #-------------

    if  dftcode == 'quantumespresso':
        parsed = parse_line(log,'charge density cutoff',0)
        if parsed is None:
            raise ValueError
        else:
            raw    = parsed.split('=')[1][:7]
            return round(ryd_to_ev * float(raw))
    else:
        return None



from typing import Dict
from dbgen.core.parsing import parse_line

def parse_pw_qe(log:str) -> int:
    """
    files   - a dictionary mapping file names to file content
    """
    # Constants
    #----------
    RYDBERG_TO_EV = 13.60569

    # Parsing
    #-------
    parsed      = parse_line(log,'kinetic-energy cutoff',0)
    if parsed is None:
        raise ValueError
    else:
        raw_rydberg = parsed.split()[-2]
        raw_num     = RYDBERG_TO_EV * float(raw_rydberg)
        return int(round(float(raw_num)))



from typing import Dict
from dbgen.core.parsing import parse_line

def parse_pw_gpaw(log:str) -> int:
    """
    Parse a planewave cutoff from GPAW logfile, bearing in mind its default values
    """

    # Parsing
    #-------
    rawmode = parse_line(log,'mode',0)
    if rawmode is None:
        raise ValueError('could not find what mode GPAW was run in')
    elif 'lcao' in rawmode:
        return 0 # non-planewave basis
    else:
        parsed  = parse_line(log,'ecut',0)
        if parsed is None:
            return 340 # gpaw default for PW mode
        else:
            raw_num = parsed.split()[-1].replace(',','')
            return int(round(float(raw_num)))



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_econv_vasp(outcar : str) -> Optional[float]:
    """
    Electronic energy convergence from INCAR
    """


    parsed = parse_line(outcar,'EDIFF',0)
    if parsed is None:
        raise ValueError('malformed OUTCAR? no "EDIFF" found')
    else:
        raw    = parsed.split('=')[1].strip().split()[0]
        return float(raw)



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_nmix_qe(log   : str
            ) -> Optional[float]:
    """
    Need a better solution eventually, some log files don't have mixing_ndim?
    """

    parsed = parse_line(log,'number of iterations used')
    if parsed is None:
        raise ValueError('Malformed QE logfile? cannot find "number of iterations used"')
    else:
        raw    = parsed.split("=")[1].strip().split()[0]
        return int(raw)



from os         import environ,system,remove,listdir
from subprocess import getstatusoutput

def get_file(pth:str)->str:
    """
    Reads contents of file
    """
    user = environ['SHERLOCK_USERNAME']
    ssh  = 'ssh %s@login.sherlock.stanford.edu '%user

    if isinstance(pth,str):
        status,content = getstatusoutput(ssh+'cat %s'%pth)
        if status == 0:
            return content
        else:
            print('could not find '+pth)
            return ''
    elif isinstance(pth,list):
        delim = "&&&&&"
        cmd = (' echo "%s"; '%delim).join(['cat %s;'%x for x in pth])
        #print('command = '+cmd)
        _,content = getstatusoutput(ssh+cmd)
        #print('content = '+content)
        files = content.split(delim)
        return ['' if 'No such file' in f else f for f in files]



# External Modules
from typing                import Any,Type,TupleMeta,CallableMeta # type: ignore
from typing                import Union as U,Callable as Call,TypeVar as TV
from json                  import JSONEncoder,dumps
from pprint                import pformat
from ast                   import parse,Lambda,walk
from os                    import linesep
from sys                   import version_info
from importlib.util        import spec_from_file_location,module_from_spec
from inspect               import (getdoc,signature,isfunction,getsourcefile # type: ignore
                                  ,getmembers,getsourcelines,_empty)

from pymatgen              import Structure  # type: ignore
import ase  # type: ignore
from networkx.classes.multigraph import MultiGraph # type: ignore

# Iternal Modules
from dbgen.support.utils import sqlite_execute,mkInsCmd,namespaceCmd
from dbgen.support.datatypes.datatypes import (DataType,Tuple,List,Dict,AnyType
                                              ,Base,NoneType,TypeVar,Union
                                              ,Callable)


################################################################################
# MOVE THIS TO UTILS IF WE END UP NEEDING TO SERIALIZE ANYTHING ELSEWHERE
#-----------------------------------------------------------------------
class Enc(JSONEncoder):
    """
    Use arbitrary serialization methods, if they're provided. Currently, this
        is used to serialize Python datatypes in a way that can be parsed into
        dbgenHaskell's "DataType" (via Aeson.fromJSON).
    """
    def default(self, o : Any)->dict:
        if hasattr(o,'toJSON'):
            return o.toJSON()
        else:
            return JSONEncoder.default(self,o)
################################################################################
version   = version_info[1]
assert version > 5, "Stop using old python3 (need 3.x, x > 5)"
class Func(object):
    """
    A function that can be used during the DB generation process.
    It is specified by:
        - a 'handle' (which is referred to in the "template" field
            of Rule objects)
        - a (full) path to a file that contains ONLY that function (at the top level)
    """
    def __init__(self,func : Call) -> None:

        # Determine if a path was given or a function
        # We'll slowly deprecate the use of paths
        #-----------------------------------------
        assert callable(func), 'Need to initialize function with a callable'

        # Extract basic info from function object
        #----------------------------------------
        self.func = func
        self.name = func.__name__           # type: ignore
        self.doc  = getdoc(func)
        self.src  = self.get_source(func)  # type: ignore
        sig       = signature(func)         # type: ignore
        self.nIn  = len(sig.parameters)

        # Determine # of outputs
        #----------------------
        output    = sig.return_annotation
        if str(output)[:11]=='typing.List':
            self.nOut = len(output.__args__)
        elif isinstance(output,TupleMeta): # type: ignore
            self.nOut = len(output.__args__)
        else:
            self.nOut = 1

        # Determine types
        #-----------------
        self.iT = [self.get_datatype(x.annotation) for x in sig.parameters.values()]
        self.oT = self.get_datatype(sig.return_annotation)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __hash__(self)->int:
        return hash(self.name)

    def apply(self,*args : Any)->Any:
        return self.func(*args) # type: ignore

    def add_func(self
                ,sqlite_pth : str
                ) -> None:
        """
        Performs analysis of PYTHON functions
        """


        cols = ['name','source','docstring'
               ,'inTypes_json','outType_json','inTypes','outType','n_in','n_out']

        binds= [self.name,self.src,self.doc
               ,dumps(self.iT,indent=4,cls=Enc)
               ,dumps(self.oT,indent=4,cls=Enc)
               ,str(self.iT),str(self.oT),self.nIn,self.nOut]
        try:
            sqlite_execute(sqlite_pth,mkInsCmd('func',cols,sqlite=True),binds)
        except Exception as e:
            raise ValueError('error with func "%s": '%self.name + str(e))
        sqlite_execute(sqlite_pth,namespaceCmd,[self.name,'Func'])


    @staticmethod
    def path_to_func(pth:str) -> Callable:
        """
        Assumes we have files with one sole function in them
        """
        spec = spec_from_file_location('',pth)
        if spec.loader is None:
            raise ValueError(pth)
        else:
            mod  = module_from_spec(spec)
            spec.loader.exec_module(mod)

            def check(o:Callable)->bool:
                return isfunction(o) and getsourcefile(o)==pth

            funcs = [o for o in getmembers(mod) if check(o[1])]

            assert len(funcs)==1,"Bad input file %s has %d functions, not 1"%(pth,len(funcs))

            return funcs[0][1]

    @staticmethod
    def get_source(f:Callable)->str:
        """Return the source code, even if it's lambda function.
        Source: http://xion.io/post/code/python-get-lambda-code.html
        """
        try:
            source_lines, _ = getsourcelines(f)
        except (IOError, TypeError):
            raise ValueError

        # Skip `def`-ed functions and long lambdas
        if len(source_lines) > 1:
            return linesep.join(source_lines).strip()

        source_text = linesep.join(source_lines).strip()

        # Find the AST node of a lambda definition so we can locate it in the source code
        source_ast = parse(source_text)
        lambda_node = next((node for node in walk(source_ast)
                            if isinstance(node, Lambda)), None)

        if lambda_node is None:  # could be a single line `def fn(x): ...`
            raise ValueError

        # HACK: Source will have some trailing junk after their definition.
        # Unfortunately, AST nodes only keep their _starting_ offsets from the original source, so we have to determine the end ourselves. We do that by gradually shaving extra junk from after the definition.
        lambda_text = source_text[lambda_node.col_offset:]
        lambda_body_text = source_text[lambda_node.body.col_offset:]
        min_length = len('lambda:_')  # shortest possible lambda expression
        while len(lambda_text) > min_length:
            try:
                # What's annoying is that sometimes the junk even parses,
                # but results in a *different* lambda. You'd probably have to  be deliberately malicious to exploit it but here's one way:
                # >bloop = lambda x: False, lambda x: True
                # >get_short_lamda_source(bloop[0])
                # Ideally, we'd just keep shaving until we get the same code, but that most likely won't happen because we can't replicate the exact closure environment.
                code = compile(lambda_body_text, '<unused filename>', 'eval')
                # Thus the next best thing is to assume some divergence due to e.g. LOAD_GLOBAL in original code being LOAD_FAST in
                # the one compiled above, or vice versa. But the resulting code should at least be the same *length*
                # if otherwise the same operations are performed in it.
                if len(code.co_code) == len(getattr(f,'__code__').co_code):
                    return lambda_text
            except SyntaxError:
                pass
            lambda_text      = lambda_text[:-1]
            lambda_body_text = lambda_body_text[:-1]

        raise ValueError

    @staticmethod
    def get_datatype(t:Type)->DataType:
        """
        Convert a Python Type object into a simpler ADT
        """
        strt = str(t)
        if   t == float:            return Base("Float")
        elif t == int:              return Base("Int")
        elif t == str:              return Base("Str")
        elif t == bool:             return Base("Bool")
        elif t == type(None):       return NoneType()

        elif t in [ase.atoms.Atoms,ase.Atoms,'Atoms']:
            return Base("Atoms")
        elif t in [Structure,'Structure']:
            return Base("Atoms")
        elif strt in ['BULK',"<class 'bulk_enumerator.bulk.BULK'>"]:
            return Base("BULK")
        elif t == MultiGraph:
            return Base("MultiGraph")

        elif str(t.__class__)==str(Union) or strt[:12]=='typing.Union':
            if version > 5:
                return Union([Func.get_datatype(x) for x in t.__args__])
            else:
                return Union([Func.get_datatype(x) for x in t.__union_set_params__])

        elif isinstance(t,TupleMeta): # type: ignore
            if version > 5:
                return Tuple([Func.get_datatype(x) for x in t.__args__])
            else:
                return Tuple([Func.get_datatype(x) for x in t.__tuple_params__])

        elif isinstance(t,CallableMeta): # type: ignore
            from_args = [Func.get_datatype(x) for x in t.__args__[:-1]]
            to_arg    = Func.get_datatype(t.__args__[-1])
            return Callable(from_args,to_arg)

        elif isinstance(t,TV): # type: ignore
            return TypeVar(t.__name__)

        elif strt[:11]=='typing.List':
            return List(Func.get_datatype(t.__args__[0]))

        elif strt[:11]=='typing.Dict':
            a1,a2 = [Func.get_datatype(x) for x in t.__args__]
            return Dict(a1,a2)

        elif strt == 'typing.Any' or t == _empty: # type: ignore
            return AnyType()

        else:
            print('NEW DATATYPE FOUND')
            import pdb;pdb.set_trace()
            raise NotImplementedError()

    @staticmethod # NEED TO LOAD FROM META.DB
    def dict2DT(x : dict)->DataType:
        tag = x.get('tag')
        if tag == 'Base':
            return Base(x['contents'])
        elif tag == 'Tuple':
            return Tuple([Func.dict2DT(y) for y in x['contents']])
        else:
            raise NotImplementedError('new case for dict2DT: ',x)

###############
# Spceial Funcs
#--------------
identity = Func(lambda x: x)



from typing import Dict,List,Tuple,Optional
from ase import Atoms # type: ignore
from string import digits
from os.path import join,exists
from os import listdir

from dbgen.core.parsing import parse_line,btw
from dbgen.core.numeric import roundfloat
from dbgen.core.lists   import flatten

ForceTuple = Tuple[int,int,Optional[float],Optional[float],Optional[float]]

def get_traj(stordir:str)->List[Tuple[Atoms,int,float,List[ForceTuple],int]]:
    """
    Returns list of triples to represent the trajectory of a structure during
    an optimization. The ASE structure, the step number along the trajectory,
    and the energy. These are parsed from a Quantum Espresso log file.
    """

    # Constants
    ryd_to_ev = 13.60569

    def mkAtom(rawatom : str) -> Tuple[str,Tuple[float,float,float]]:
        _,sym,_,_,_,_,x,y,z,_ = rawatom.split()
        remove_digits = str.maketrans('', '', digits)
        clean_symb = sym.translate(remove_digits)
        return (clean_symb,(roundfloat(x),roundfloat(y),roundfloat(z)))

    def parse_cell_line(l : str) -> List[float]:
        return [roundfloat(x) for x in l.split()[3:-1]]

    def parse_force_line(step : int
                        ,l    : str
                        ) -> ForceTuple:
        _,ind1,_,_,_,_,fx,fy,fz = l.split()
        return (step,int(ind1)-1,float(fx)* ryd_to_ev,float(fy)* ryd_to_ev,float(fz)* ryd_to_ev)

    def parse_section(sect    : str
                     ,stepnum : int
                     ) -> Tuple[Atoms,int,float,List[ForceTuple]]:

        cell_text,cell_ind = btw(sect,'crystal axes:','reciprocal')
        cell_lines = cell_text.strip().split('\n')[1:]
        cell = [parse_cell_line(x) for x in cell_lines[:3]]

        a_text,a_ind  = btw(sect,"site n.","number of k points",cell_ind)
        rawatoms = a_text.strip().split('\n')[1:]
        allatoms = list(map(mkAtom,rawatoms))
        atoms    = Atoms(symbols   = [a[0] for a in allatoms]
                        ,positions = [a[1] for a in allatoms]
                        ,cell      = cell)
        atoms.wrap()


        eng_line = parse_line(sect[a_ind:],'total energy',0)
        ts_line  = parse_line(sect[a_ind:],'smearing contrib',-1)
        if eng_line is None:
            for line in reversed(sect.split('\n')):
                if 'total energy' in line:
                    eng_line = line
                elif 'smearing contrib' in line:
                    ts_line = line

        e        = float(eng_line.split()[-2])
        ts       = float(ts_line.split()[-2])
        ev_eng   = (e-ts/2) * ryd_to_ev

        end_pattern = 'Total force' if 'Total force' in sect[a_ind:] else '$'
        f_text,_ =  btw(sect,'Forces acting on',end_pattern,a_ind)
        force_lines = f_text.strip().split('\n')[2:]


        forces = [parse_force_line(stepnum,x) for x in force_lines
                if len(x.split())==9 and 'force =' in x]


        if len(forces)!=len(atoms):
            forces = [(stepnum,i,None,None,None) for i in range(len(atoms))]

        return (atoms,stepnum,ev_eng,forces,0)


    # Main program
    #--------------

    # Get log file
    #-------------
    if exists(join(stordir,'log')):
        log = join(stordir,'log')
    else:
        for subdir in listdir(stordir):
            if exists(join(stordir,subdir,'log')):
                log = join(stordir,subdir,'log')
    with open(log,'r') as f: logfile = f.read()

    sections = logfile.split('JOB DONE') # re.split('JOB DONE',
    if len(sections)>5:
        sections = sections[-5:]
    output = [parse_section(x,n) for n,x in enumerate(sections)
                if 'total energy' in x and 'reciprocal' in x] # sometimes there are malformed sections that are missing key pieces of info ... discard them? HACK

    output[-1] = tuple([*output[-1][:-1],1]) # set final for finaltraj
    return output



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_econv_gpaw(log   : str
             ) -> Optional[float]:
    """
    Electronic energy convergence
    """
    init = log.split('Initialize')[0]
    parsed = parse_line(init,'energy:',0)
    if parsed is None: # the first hit for 'energy:' is NOT in the input parameter section
        return 0.0005 # default
    else:
        raw    = parsed.split('energy:')[-1]
        pure   = raw.replace('}','')
        return float(pure)



from typing import Dict

from dbgen.core.parsing import parse_line

def parse_xc_qe(log : str) -> str:
    """
    Finds the exchange correlation functional
    """
    parsed_line = parse_line(log,'Exchange-correlation',0)
    if parsed_line is None:
        raise ValueError
    else:
        raw_val       = parsed_line.split('=')[1].split()[0]
        processed_val = raw_val.replace(',','').replace("'",'')

        return  processed_val.strip()



# External Modules
from typing import Tuple
import subprocess,os
from os.path import join
from json import loads,dumps
from subprocess import getstatusoutput
# Internal Modules
from dbgen.scripts.IO.get_file import get_file
################################################################################
def get_job_metadata(stordir : str
                    ,code    : str
                    ,raw_pdict : str
                    ) -> Tuple[str,int,str,str,str,str]:
    """
    Takes a path to a DFT calculation and extracts information from runtime.json
    Also keeps a record of where the information was taken from.
    """
    try:
        runtime   = loads(get_file(join(stordir,"runtime.json")))
        pdict     = {k:v for k,v in loads(raw_pdict).items()
                     if 'pckl' not in k and 'project' not in k} # discard binary
        jobkind   =  pdict.get('jobkind','<no jobkind>')

        if 'kwargs' in pdict.keys():
            kwargs  = loads(pdict['kwargs'])
            jobname = kwargs.get('job_name','<no jobname>')
        else:
            jobname =  pdict.get('job_name')
            if jobname is None:
                jobname =  pdict.get('name','<no jobname>')

        return (runtime.get('user') # type: ignore
               ,runtime.get('timestamp')
               ,runtime.get('working_directory')
               ,jobkind
               ,jobname) # type: ignore

    except IOError: # job NOT produced by CataLog
        user = os.environ['SHERLOCK_USERNAME']
        ssh  = 'ssh %s@login.sherlock.stanford.edu '%user

        usr    = getstatusoutput(ssh+'stat -c %U {}'.format(stordir))[1]
        tstamp = getstatusoutput(ssh+'stat -c %Y {}'.format(stordir))[1]

        if code=='chargemol':
            jk = 'chargemol'
        else:
            jk = 'relax'
        # WE DON'T SAVE ANYTHING TO IDENTIFY WHETHER JOB IS VIBJOB!!!
        return (usr,int(tstamp),stordir,jk,stordir) # type: ignore

if __name__=='__main__':
    import sys
    print(get_job_metadata(sys.argv[1],'',''))



from typing import List,FrozenSet
from dbgen.support.datatypes.relation   import Relation
from dbgen.support.datatypes.object     import Object,Schema
from dbgen.support.datatypes.func       import Func
from dbgen.support.datatypes.dbg        import DBG
from dbgen.support.datatypes.misc       import ConnectInfo

################################################################################
class Model(object):
    def __init__(self
                ,schema   : Schema
                ,relations : List[Relation]
                ) -> None:
        self.schema    = schema
        self.relations = relations
        self.objects   = schema.objects

    def createDBG(self)->DBG:
        tables = [o._mkTable(self.schema) for o in self.objects]
        rules  = [r.mk_rule() for r in self.relations]
        return DBG(tables=tables,rules=rules)

    def run(self
           ,reset   : bool = False
           ,catalog : bool = False
           ) -> None:
        db = ConnectInfo()
        x  = set() if catalog else {'catalog'}
        self.createDBG().run_all(db,reset=reset,xclude=x,parallel=False)



from typing     import List,Tuple
from subprocess import getstatusoutput
from json       import loads
from os         import environ
def get_chargemol_logfiles(rootpath : str
                    #      ,filejson : str
                          ) -> Tuple[List[str],List[str]]:
    """
    Searches for chargemol calculation logfiles.
    Returns unzipped (stordir,logfile) pairs

    Requirements to be flagged as a DFT job:
        - There must exist at least one .traj file in the directory
        - If there's a DDEC analysis file, we'll assume it's a Chargemol job
            (need to check if it is completed)

    """
    ############################
    # Initialize Variables
    #---------------------
    dirs,logs  = [],[]
    user       = environ['SHERLOCK_USERNAME']

    # Main Program
    #-------------
    ssh = 'ssh %s@login.sherlock.stanford.edu '%user
    cmd = """for f in $(find {0} -name total_cube_DDEC_analysis.output);
                    do
                        result=$(tail -3 $f | grep -c "Finished chargemol") ;
                if [ ! $result -eq 0 ]; then
                    echo "$f"
                    fi
                done""".format(rootpath)
    exit,output = getstatusoutput(ssh+cmd)
    assert exit==0,'Failure in get_chargemol_logfile bash execution: '+output
    for logfile in output.split('\n'):
        if logfile:
            last_slash = logfile.rfind('/')
            dir = logfile[:last_slash]
            if True: #dir not in files:
                dirs.append(dir)
                logs.append(logfile)

    return dirs,logs



from typing import Tuple
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer       # type: ignore
import numpy as np                                              # type: ignore
from pymatgen import Lattice,Structure                          # type: ignore
from pymatgen.core.surface import get_recp_symmetry_operation   # type: ignore
from monty.fractions import gcd_float                           # type: ignore
from ase.visualize                      import view             # type: ignore
from pymatgen.io.ase                    import AseAtomsAdaptor  # type: ignore


def get_miller(structure:Structure)->Tuple[int,int,int]:
    """
    Courtesy of Joey M.
    """
    # Get normal uvec
    normal = np.cross(structure.lattice.matrix[0], structure.lattice.matrix[1])
    normal /= np.linalg.norm(normal)

    # Get the conventional lattice in the same basis as the current structure
    sga = SpacegroupAnalyzer(structure,symprec=0.15)
    trans_mat = sga.get_symmetry_dataset()['transformation_matrix']
    conv_latt = Lattice(np.transpose(np.dot(np.transpose(
                structure.lattice.matrix), np.linalg.inv(trans_mat))))

    # Convert to frac coords, then normalize to gcd
    normal_in_frac = conv_latt.get_fractional_coords(normal)
    miller = np.array([v if abs(v) > 1e-8 else 0 for v in normal_in_frac])
    miller /= abs(gcd_float([m for m in miller if not np.isclose(m, 0)]))
    miller = miller.round(5)

    # If you really want, you can do this thing and it'll sort all of the
    # equivalent miller indices and give the last one in a sort by value
    recip_symmops = get_recp_symmetry_operation(sga.get_conventional_standard_structure())
    equiv_millers = [tuple([int(x) for x in op.operate(miller)]) for op in recip_symmops]
    miller = sorted(equiv_millers)[-1]

    return miller

if __name__=='__main__':
    import sys
    from ase.io import read # type: ignore
    from dbgen.scripts.Pure.Atoms.surface_bulkstruct import surface_bulkstruct  # type: ignore
    pth = sys.argv[1]
    print(get_miller(surface_bulkstruct(read(pth))))



from typing import Dict,Optional
import json
from dbgen.core.parsing import parse_line

def get_fmax(dftcode     : str
            ,paramsdict  : str
            ,outcar      : str
            ) -> Optional[float]:
    """
    docstring
    """
    if dftcode == 'vasp':
        return -float(parse_line(outcar,'EDIFFG').split('=')[-1])
    else:
        params = json.loads(paramsdict)
        return params.get('fmax')     # no conceivable way of getting this from log



# External Modules
from typing import Tuple,List,TYPE_CHECKING
if TYPE_CHECKING:
    from ase import Atoms       # type: ignore
    from pymatgen import Site   # type: ignore
from pymatgen.analysis.structure_analyzer import VoronoiConnectivity  # type: ignore
from pymatgen.io.ase                      import AseAtomsAdaptor  # type: ignore
from math import floor
from json import dumps
import numpy as np # type: ignore

# Internal Modules
from dbgen.core.lists import flatten

def solid_angle(atoms : 'Atoms'
        ,tol   : float   = 1 # MIN SOLID ANGLE
        ) -> str:
    """Take an Atoms object and return bonddata USING solid angles """

    positions = atoms.get_positions()
    floors    = np.vectorize(floor)

    pmg  = AseAtomsAdaptor().get_structure(atoms)
    vc   = VoronoiConnectivity(pmg)
    arr  = vc.connectivity_array
    output = []
    for i in range(len(atoms)):
        for j in range(i,len(atoms)):
            print('i = %d,j = %d'%(i,j))
            imgs = np.where( arr[i][j] > tol )
            angles = arr[i][j][imgs]
            offs   = vc.offsets[imgs]
            for a,(x,y,z) in zip(angles,offs):
                print((i,j,a,x,y,z))
                output.append((i,j,a,x,y,z))
    return dumps(output)

if __name__=='__main__':
    import sys
    from ase.io import read # type: ignore
    if len(sys.argv)>1:
        pth = sys.argv[1]
    else:
        pth = '/Users/ksb/dbgen/data/suncatdata/percentiles/0/qn.traj'
    print(solid_angle(read(pth)))



from typing import TypeVar,Callable
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')

from dbgen.core.misc import identity
################################################################################
def dftcode_dependent(dftcode      : str
                     ,raw_input    : A
                     ,gpaw_func    : Callable[[B],C]
                     ,qe_func      : Callable[[B],C]
                     ,vasp_func    : Callable[[B],C]
                     ,pre_process  : Callable[[A],B] = identity
                     ,post_process : Callable[[C],D] = identity
                     ) -> D:
    """
    Higher order function for common paradigm: three completely different
    functions needed depending on DFT code
    """
    input = pre_process(raw_input)

    if dftcode == 'gpaw':
        raw_output =  gpaw_func(input)
    elif dftcode == 'quantumespresso':
        raw_output =  qe_func(input)
    elif dftcode == 'vasp':
        raw_output =  vasp_func(input)
    else:
        raise NotImplementedError('No implementation for dftcode ',dftcode)


    return post_process(raw_output)




#Typing imports
import typing as typ

#External imports
from os             import environ
from argparse       import ArgumentParser
from pprint         import pformat
from distutils.util import strtobool
from ast            import literal_eval
from json           import load

# Internal Modules
from dbgen.support.utils          import ConnectInfo
from dbgen.core.misc              import levenshteinDistance
from dbgen.inputs.development     import dbg
"""
Command line interface for executing a database update

Requires DB_JSON environment variable to specify database that will be modified

The following flags control program behavior:
    - verbose
    - local
    - reset
    - add
    - meta
    - only
    - xclude
    - jobs
    - catalog
    - test
    - parallel

"""
#Internal Imports
import parser # type: ignore

class CLICommand:
    short_description = "Command line interface for executing a database update"
    description = ("""
The following flags control program behavior:
        - verbose
        - local
        - reset
        - add
        - meta
        - only
        - xclude
        - jobs
        - catalog
        - test
        - parallel""")

    @staticmethod
    def add_arguments(parser : 'parser') -> None: # type: ignore
        parser.add_argument('--verbose'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = """Leave it off please - hard to find a reasonable
                                        amount of things that's neither too little nor
                                        too much ... maybe the input for this should
                                        actually be 'high 'low' or 'medium'?""")

        parser.add_argument('--local'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Default: if on sherlock, don\'t run local')

        parser.add_argument('--reset'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Nuke the DB first - needed if you make schema changes')

        parser.add_argument('--add'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Try to add columns (if you make benign schema additions)')

        parser.add_argument('--meta'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Make a meta.db')

        parser.add_argument('--only'
                           ,default = ''
                           ,help    = 'Run only the (space separated) rules')

        parser.add_argument('--xclude'
                           ,default = 'underlying ads_triple'
                           ,help    = 'Do not run the (space separated) rules')

        parser.add_argument('--jobs'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Run all job_loading rules (takes time to rerun even if no new jobs)')

        parser.add_argument('--catalog'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Run only the catalog job loading rule ')

        parser.add_argument('--test'
                           ,default = ''
                           ,nargs   = 2
                           ,help    = 'Give a Tranform Name and an input dictionary (as parsable string) to test')

        parser.add_argument('--parallel'
                           ,default = True
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Parallelize processing of multiple Query inputs. (Turn off if using pdb)')

    @staticmethod
    def run(args : typ.Any) -> None:
        """
        Executes DB modifying procedure described by a DBG instance
        """
        print('args are: \n'+pformat(args))


        if args.test:
            # Don't interact with database
            #------------------------------------
            dic = literal_eval(args.test[1])
            assert isinstance(dic,dict), "We need a dictionary to test a Rule's input"
            dbg.test_input(args.test[0],dic)
        else:
            # Actually interact with database
            #------------------------------------
            if args.local:
                db = ConnectInfo()
            else:
                with open(environ['DB_JSON'],'r') as f:
                    db = ConnectInfo(**load(f))

            if args.meta:
                dbg.test_meta() # create a meta-db

            # Determine which job-loading rules to include
            #------------------------------------------------------------
            if not args.jobs:
                args.xclude+=''' stray_gpaw_logfiles
                                 stray_qe_logfiles
                                 stray_chargemol_logfiles ''' # these T's ONLY run if --jobs

            if (not args.reset) and (not (args.catalog or args.jobs)):
                args.xclude+=''' catalog ''' # this run if EITHER --jobs OR --catalog...provided we didn't just nuke DB

            # Process string inputs
            #------------------------
            only   = set(args.only.split())
            xclude = set(args.xclude.split()) - only

            # Verify every  rule excluded/included by name actually exists
            # if first 5 chars match or within edit distance of 5, suggest alternatives
            #------------------------------------------------------------------------------------
            for w in (only | xclude):
                match = False
                close = []
                for n in [t.name for t in dbg.rules]:
                    d = levenshteinDistance(w,n)
                    upW,upN = max(len(w),5), max(len(n),5) # variables for safe indexing
                    if d == 0:
                        match = True
                        break
                    elif d < 5 or w[:upW] == n[:upN]:
                        close.append(n) # keep track of near-misses
                if not match:
                    did_you = "Did you mean %s"%close if close else ''
                    raise ValueError("No match found for %s\n%s"%(w,did_you))

            dbg.run_all(db
                       ,only    = only
                       ,xclude  = xclude
                       ,reset   = args.reset
                       ,add     = args.add
                       ,verbose = args.verbose
                       ,parallel= args.parallel)



from typing import List
import numpy as np  # type: ignore
import ase          # type: ignore

def get_system_type(atoms : ase.Atoms) -> str:
    """
    Helpful docstring
    """
    ############################################################################
    # Parameters
    #------------
    cutoff    = 6  # A
    cell_abc  = list(map(np.linalg.norm,atoms.get_cell()[:3]))

    big_boy   = atoms.copy()*(2,2,2)
    xs,ys,zs  = zip(*big_boy.positions)
    sx,sy,sz  = [sorted(q) for q in [xs,ys,zs]]
    gap       = [max(np.diff(s)) for s in [sx,sy,sz]]
    pbc       = [gap[i] < cutoff for i in range(3)]

    if   all(pbc): return 'bulk'
    elif any(pbc): return 'surface'
    else:          return 'molecule'

if __name__=='__main__':
    import sys
    from dbgen.core.misc import anytraj
    a = anytraj(sys.argv[1])
    print(get_system_type(a))



from typing import Dict,Optional
from dbgen.core.parsing import parse_line

def get_econv_qe(log : str
             ) -> Optional[float]:
    """
    Electronic energy convergence from pw.inp file
    """

    # Constants
    #----------
    ryd_to_ev = 13.60569

    # main program
    #------------
    line = parse_line(log,'convergence threshold',0)
    if line is None:
        raise ValueError('malformed QE logfile ? no "convergence threshold" found')
    else:
        raw = line.split('=')[-1]
        return round(ryd_to_ev * float(raw),6)



from typing import Dict,Optional

from dbgen.core.parsing import parse_line

def get_cell_factor(pwinp : str
                   ,dftcode : str
                   ) -> Optional[float]:
    """
    Get cell factor for vc relax jobs
    """
    if dftcode !='quantumespresso':
        return None

    line  = parse_line(pwinp,'cell_factor=')
    if line is None:
        return None
    else:
        raw = line.replace('d','=').split('=')[1]
        return float(raw)

if __name__=='__main__':
    # Test on a QE pw.inp file
    import sys
    with open(sys.argv[1],'r') as f:
        print(get_cell_factor(f.read(),'quantumespresso'))



from typing import Tuple,List,Dict
from gpaw import GPAW  #type: ignore
from gpaw.utilities import h2gpts #type: ignore
from ase.io import read,write     #type: ignore
from ase import Atoms,Atom       #type: ignore
from ase.units import Bohr   #type: ignore
import os
from string import ascii_uppercase,digits
from random import choices
from collections import defaultdict
import numpy as np   #type: ignore
from json import dumps


def quickgraph(atoms:Atoms)->str:
    calc = GPAW(mode='lcao',basis = 'dzp'
                    ,gpts          = h2gpts(0.15, atoms.get_cell(), idiv=8)
                    ,txt           = 'log'
                    ,fixdensity    = True
                    ,convergence   = {'energy':float('inf')
                                     ,'density':float('inf')})

    stordir = ''.join(choices(ascii_uppercase+digits, k=8))
    os.chdir(stordir)
    atoms.set_calculator(calc)
    atoms.get_potential_energy()
    density = calc.get_all_electron_density() * Bohr**3
    write(stordir+'/total_density.cube',atoms,data=density)
    job_control = '\n'.join(['<net charge>',"0.0","</net charge>"
                            ,"<periodicity along A, B, and C vectors>"
                            ,".true.",".true.",".true."
                            ,"</periodicity along A, B, and C vectors>"
                            ,'<compute BOs>','.true.','</compute BOs>'
                            ,'<atomic densities directory complete path>'
                            ,os.environ['CHARGEMOL_DENSITIES']
                            ,'</atomic densities directory complete path>'
                            ,'<charge type>','DDEC6','</charge type>'])

    os.system(os.environ['CHARGEMOL_BINARY'])

    ###############################################################
    def parse_line(line : str) -> Tuple[int,float,List[int]]:
        """
        Get bonded atom, bond order, and offset
        """
        assert line[:16]==' Bonded to the (', "No parse line ->"+line+'<- %d'%len(line)
        offStr,indStr,boStr = line[16:].split(')')
        offset = [int(x) for x in offStr.split(',')] # Chunk containing offset info
        ind    = int(indStr.split()[-3]) - 1        # Chunk containing index info (chargemol starts at 1, not 0)
        bo     = float(boStr.split()[4])            # Chunk containing B.O. info
        return (ind,bo,offset)

    class PotentialEdge(object):
        """
        Container for information we need to decide later if it's a
        graph-worthy bond. This class is nothing more than a dictionary.
        """
        def __init__(self
                    ,fromNode   : int
                    ,toNode     : int
                    ,bond_order : float
                    ,offset     : List[int]
                    ) -> None:

            self.fromNode = fromNode
            self.toNode   = toNode
            self.bondorder= bond_order
            self.pbc_x    = offset[0]
            self.pbc_y    = offset[1]
            self.pbc_z    = offset[2]


    class BondOrderSection(object):
        """
        Process one section of the Bond Order output of Chargemol
        """
        def __init__(self
                    ,ind       : int
                    ,raw_lines : List[str]
                    ,pbcdict   : dict
                    ) -> None:

            self.ind      = ind
            self.bonds    = [parse_line(x) for x in raw_lines]
            self.pbcdict  = pbcdict

        def _relative_shift(self
                           ,i : int
                           ,j : int
                           ) -> np.array:
            """
            Given a pbc_dict and two indices, return the original pbc shift
            for a bond from i to j
            """
            pi,pj = [np.array(self.pbcdict[x]) for x in [i,j]]
            return pj - pi

        def makeEdge(self
                    ,tup :  Tuple[int,float,List[int]]
                    ) -> PotentialEdge:
            """
            Creates an Edge instance from the result of a parsed Bond Order log line
            """
            (toInd,bo,offset) = tup
            fromInd = self.ind

            # correct for WRAPPING atoms
            offset = (np.array(offset) +  self._relative_shift(fromInd,toInd)).tolist()

            shift = np.dot(offset, atoms.get_cell()) # PBC shift
            p1    = atoms[fromInd].position
            p2    = atoms[toInd].position + shift
            d     = np.linalg.norm(p2-p1)
            return PotentialEdge(fromInd,toInd,bo,offset)

        def make_edges(self)->List[PotentialEdge]:
            """Apply edgemaker to result of parsing logfile lines"""
            return  [self.makeEdge(b) for b in self.bonds]



    def mk_pbc_dict(atoms : Atoms
                   ) -> Dict[int,Tuple[int,int,int]]:
        """
        Helpful docstring
        """
        def g(tup : tuple)->Tuple[int,int,int]:
            """
            Helper function to yield tuples for pbc_dict
            """
            def f(x : float)->int:
                """
                Helper function for g
                """
                if   x < 0: return -1
                elif x < 1: return 0
                else:       return 1

            x,y,z = tup
            return (f(x),f(y),f(z))

        scaled_pos  = atoms.get_scaled_positions(wrap=False).tolist()
        scaled_pos_ = zip(range(len(atoms)),scaled_pos)
        pbc_dict    = {i : g(p) for i,p in scaled_pos_}
        return pbc_dict

    def parse_chargemol_pbc(header_lines : List[str]
                           ,cell         : List[float]
                           ) -> Dict[int,Tuple[int,int,int]]:
        """
        Helpful docstring
        """
        atoms = Atoms(cell=cell)

        for i,l in enumerate(header_lines[2:]):
            try:
                if '*' in l:
                    l = l.replace('*','') + ' ???' #Asterisks are touching the z value
                s,x,y,z,_ = l.split()
                p = [float(q) for q in [x,y,z]]
                atoms.append(Atom(s,position=p))
            except Exception as e:
                pass #print('exception???',e)

        return mk_pbc_dict(atoms)
    def dict_diff(d1 : Dict[int,np.array]
                 ,d2 : Dict[int,np.array]
                 ) -> Dict[int,np.array]:
        """
        Helpful docstring
        """
        return {i: np.array(d2[i]) - np.array(d1[i]) for i in d1.keys()}


    header,content    = [], [] # type: Tuple[List[str],List[str]]
    sections          = []     # type: List[BondOrderSection]
    head_flag,counter = True,-1       # Initialize
    filepath = stordir+'/DDEC6_even_tempered_bond_orders.xyz' # File to parse

    pbcdict = mk_pbc_dict(atoms) # the 'true' PBC coordinate of each atom

    filtedges = []
    with open(filepath,'r') as f:
        for line in f:
            if not (line.isspace() or line[:4] == ' 201'): # remove blank lines and calendar date
                if line[1]=='=':
                    head_flag = False                                            # no longer in header  section
                    chargemol_pbc = parse_chargemol_pbc(header,atoms.get_cell()) # PBC of chargemol atoms

                elif head_flag: header.append(line)     # we're still in the header section
                elif 'Printing' in line:                # we're describing a new atom
                    content = []                        # reset buffer content
                    counter += 1                        # update index of our from_index atom
                elif 'sum' in line:                     # summary line at end of a section
                    dic = dict_diff(pbcdict,chargemol_pbc)
                    edgs = BondOrderSection(counter,content,dic).make_edges()
                    filtedges.extend([e for e in edgs
                        if  (e.bondorder > 0.04) and (e.fromNode <= e.toNode)])
                else:
                    content.append(line)              # business-as-usual, add to buffer

    return dumps([e.__dict__.values() for e in filtedges])

if __name__ == '__main__':
    a = read('~/dbgen/data/suncatdata/ksb/150792226224/init.traj')
    print(quickgraph(a))



from typing import Dict,Tuple
import ast

from dbgen.core.parsing import parse_line

def get_kpts_vasp(trip:Tuple[str,str,str]) -> Tuple[int,int,int]:
    """
    Parse VASP kpt input from KPOINTS file
    """
    _,_,kptcar = trip
    line = kptcar.split('\n')[-2]
    raw  = [int(x) for x in line.split()]
    return tuple(raw) #type: ignore



from typing import Dict,List,Tuple,Optional
MbFloat = Optional[float]
ForceTuple = Tuple[int,int,MbFloat,MbFloat,MbFloat]
from ase import Atoms,Atom # type: ignore
from string import digits
from os.path import join,exists

from dbgen.core.parsing import parse_line,btw
from dbgen.core.numeric import roundfloat
from dbgen.core.lists   import flatten
from dbgen.core.misc    import anytraj

def get_traj(stordir : str)->List[Tuple[Atoms,int,float,List[ForceTuple],int]]:
    """
    we don't do optimizations in gpaw???

    Returns list of triples to represent the trajectory of a structure during
    an optimization. The ASE structure, the step number along the trajectory,
    and the energy. These are parsed from a GPAW log file.
    """

    # Auxilliary Functions
    #---------------------
    def gpawl2cell(l:str)->list:
        return [roundfloat(x) for x in l.split()[-5:-2]]
    def parse_force(l:str)->ForceTuple:
        i,_,fx,fy,fz = l.split()
        return (0,int(i),float(fx),float(fy),float(fz))

    # Main Program
    #--------------

    with open(join(stordir,'log'),'r') as f: log = f.read()

    # Get atoms
    #-----------
    traj = anytraj(stordir)
    atom_text,atom_ind = btw(log,'Positions','Unit cell')

    a_lines = atom_text.split('\n')[1:-2]
    symbs,pos = [],[]
    for a in a_lines:
        items = a.split()
        symbs.append(items[1])
        pos.append(list(map(float,items[2:5])))


    # Get Cell
    #---------
    cell_text,cell_ind = btw(log,'periodic','Effective grid spacing',atom_ind)
    celllines = cell_text.split('\n')[1:-2]
    cell =[gpawl2cell(x) for x in celllines]

    # Get Energy
    #-----------
    eng =float(parse_line(log[cell_ind:],'Extrapolated:',-1).split()[-1])

    # Get Forces
    #----------
    force_text,_ = btw(log,'Forces in eV/Ang:','Timing',cell_ind)
    force_lines = force_text.strip().split('\n')[1:-1]
    # Put Atoms together
    #------------------
    atoms = Atoms(symbols=symbs,positions=pos,cell=cell
                ,constraint=traj.constraints,tags=traj.get_tags())

    atoms.wrap()
    # Handle case of not finding forces
    #---------------------------------
    if len(force_lines)==len(atoms):
        forces = [parse_force(x) for x in force_lines] # <- hack
    else:
        forces = [(0,i,None,None,None) for i in range(len(atoms))]


    return  [(atoms,0,eng,forces,1)]
